# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implements a Zarr :class:`~.SyncedCollection` backend."""
from copy import deepcopy

from .. import SyncedCollection, SyncedDict, SyncedList
from ..validators import require_string_key

try:
    import numcodecs

    ZARR = True
except ImportError:
    ZARR = False


class ZarrCollection(SyncedCollection):
    r"""A :class:`~.SyncedCollection` that synchronizes with a Zarr group.

    Since Zarr is designed for storage of array-like data, this backend implements
    synchronization by storing the collection in a 1-element object array. The user
    provides the group within which to store the data and the name of the data in
    the group.

    **Thread safety**

    The :class:`ZarrCollection` is not thread-safe.

    Parameters
    ----------
    group : zarr.hierarchy.Group
        The Zarr group in which to store data.
    name : str
        The name under which this collection is stored in the Zarr group.
    codec : numcodecs.abc.Codec
        The encoding mechanism for the data. If not provided, defaults to JSON
        encoding (Default value: None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    """

    _backend = __name__  # type: ignore

    def __init__(self, group=None, name=None, codec=None, *args, **kwargs):
        if not ZARR:
            raise RuntimeError(
                "The Zarr package must be installed to use the ZarrCollection."
            )

        super().__init__(**kwargs)
        self._group = group
        self._name = name
        self._object_codec = numcodecs.JSON() if codec is None else codec

    def _load_from_resource(self):
        """Load the data from the Zarr group.

        Returns
        -------
        Collection or None
            An equivalent unsynced collection satisfying :meth:`~.is_base_type` that
            contains the data in the Zarr group. Will return None if associated
            data is not found in the Zarr group.

        """
        try:
            return self._group[self._name][0]
        except KeyError:
            return None

    def _save_to_resource(self):
        """Write the data to Zarr group."""
        data = self._to_base()
        dataset = self._group.require_dataset(
            self._name,
            overwrite=True,
            shape=1,
            dtype="object",
            object_codec=self._object_codec,
        )
        dataset[0] = data

    def __deepcopy__(self, memo):
        if self._root is not None:
            return type(self)(
                group=None,
                name=None,
                data=self._to_base(),
                parent=deepcopy(self._root, memo),
            )
        else:
            return type(self)(
                group=deepcopy(self._group, memo),
                name=self._name,
                data=None,
                parent=None,
            )

    @property
    def codec(self):
        """numcodecs.abc.Codec: The encoding method used for the data."""
        return self._object_codec

    @codec.setter
    def codec(self, new_codec):
        self._object_codec = new_codec

    @property
    def group(self):
        """zarr.hierarchy.Group: The Zarr group storing the data."""
        return self._group

    @property
    def name(self):
        """str: The name of this data in the Zarr group."""
        return self._name


class ZarrDict(ZarrCollection, SyncedDict):
    r"""A dict-like data structure that synchronizes with a Zarr group.

    Examples
    --------
    >>> doc = ZarrDict('data')
    >>> doc['foo'] = "bar"
    >>> assert doc['foo'] == "bar"
    >>> assert 'foo' in doc
    >>> del doc['foo']
    >>> doc['foo'] = dict(bar=True)
    >>> doc
    {'foo': {'bar': True}}

    Parameters
    ----------
    group : zarr.hierarchy.Group, optional
        The group in which to store data (Default value = None).
    name : str, optional
        The name of the collection (Default value = None).
    data : :class:`collections.abc.Mapping`, optional
        The initial data passed to :class:`ZarrDict`. If ``None``, defaults to
        ``{}`` (Default value = None).
    parent : ZarrCollection, optional
        A parent instance of :class:`ZarrCollection` or ``None``. If ``None``,
        the collection owns its own data (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    While the :class:`ZarrDict` object behaves like a :class:`dict`, there are important
    distinctions to remember. In particular, because operations are reflected
    as changes to an underlying database, copying (even deep copying) a
    :class:`ZarrDict` instance may exhibit unexpected behavior. If a true copy is
    required, you should use the call operator to get a dictionary
    representation, and if necessary construct a new :class:`ZarrDict` instance.

    """

    _validators = (require_string_key,)

    def __init__(self, group=None, name=None, data=None, parent=None, *args, **kwargs):
        super().__init__(
            group=group, name=name, data=data, parent=parent, *args, **kwargs
        )


class ZarrList(ZarrCollection, SyncedList):
    r"""A list-like data structure that synchronizes with a Zarr group.

    Only non-string sequences are supported by this class.

    Examples
    --------
    >>> synced_list = ZarrList('data')
    >>> synced_list.append("bar")
    >>> assert synced_list[0] == "bar"
    >>> assert len(synced_list) == 1
    >>> del synced_list[0]

    Parameters
    ----------
    group : zarr.hierarchy.Group, optional
        The group in which to store data (Default value = None).
    name : str, optional
        The name of the collection (Default value = None).
    data : non-str :class:`collections.abc.Sequence`, optional
        The initial data passed to :class:`ZarrList`. If ``None``, defaults to
        ``[]`` (Default value = None).
    parent : ZarrCollection, optional
        A parent instance of :class:`ZarrCollection` or ``None``. If ``None``,
        the collection owns its own data (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    While the :class:`ZarrList` object behaves like a :class:`list`, there are important
    distinctions to remember. In particular, because operations are reflected
    as changes to an underlying database, copying (even deep copying) a
    :class:`ZarrList` instance may exhibit unexpected behavior. If a true copy is
    required, you should use the call operator to get a dictionary
    representation, and if necessary construct a new :class:`ZarrList` instance.

    """

    def __init__(self, group=None, name=None, data=None, parent=None, *args, **kwargs):
        super().__init__(
            group=group, name=name, data=data, parent=parent, *args, **kwargs
        )



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implements the :class:`SyncedList`.

This implements a list-like data structure that also conforms to the
:class:`~.SyncedCollection` API and can be combined with any backend type to
give a list-like API to a synchronized data structure.
"""

from collections.abc import MutableSequence, Sequence

from ..numpy_utils import (
    _convert_numpy,
    _is_atleast_1d_numpy_array,
    _numpy_cache_blocklist,
)
from ..utils import AbstractTypeResolver
from .synced_collection import SyncedCollection, _sc_resolver

# Identifies sequences, which are the base type for this class.
_sequence_resolver = AbstractTypeResolver(
    {
        "SEQUENCE": (
            lambda obj: (isinstance(obj, Sequence) and not isinstance(obj, str))
            or _is_atleast_1d_numpy_array(obj)
        ),
    },
    cache_blocklist=_numpy_cache_blocklist,
)


class SyncedList(SyncedCollection, MutableSequence):
    r"""Implementation of list data structure.

    The SyncedList inherits from :class:`~synced_collection.SyncedCollection`
    and :class:`~collections.abc.MutableSequence`. Therefore, it behaves similar
    to a :class:`list`.

    Parameters
    ----------
    data : Sequence, optional
        The initial data to populate the list. If ``None``, defaults to
        ``[]`` (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    While the :class:`SyncedList` object behaves like a :class:`list`, there
    are important distinctions to remember. In particular, because operations
    are reflected as changes to an underlying backend, copying (even deep
    copying) a :class:`SyncedList` instance may exhibit unexpected behavior. If
    a true copy is required, you should use the :meth:`~.__call__` method to
    get a :class:`list` representation, and if necessary construct a new
    :class:`SyncedList`.

    """

    # The _validate parameter is an optimization for internal use only. This
    # argument will be passed by _from_base whenever an unsynced collection is
    # being recursively converted, ensuring that validation only happens once.
    def __init__(self, data=None, _validate=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if data is None:
            self._data = []
        else:
            if _validate:
                self._validate(data)
            data = _convert_numpy(data)
            with self._suspend_sync:
                self._data = [
                    self._from_base(data=value, parent=self) for value in data
                ]

    @classmethod
    def is_base_type(cls, data):
        """Check whether the data is an non-string Sequence.

        Parameters
        ----------
        data : Any
            Data to be checked

        Returns
        -------
        bool

        """
        return _sequence_resolver.get_type(data) == "SEQUENCE"

    def _to_base(self):
        """Convert the SyncedList object to a :class:`list`.

        Returns
        -------
        list
        An equivalent raw :class:`list`.

        """
        converted = []
        for value in self._data:
            switch_type = _sc_resolver.get_type(value)
            if switch_type == "SYNCEDCOLLECTION":
                converted.append(value._to_base())
            else:
                converted.append(value)
        return converted

    def _update(self, data=None, _validate=False):
        """Update the in-memory representation to match the provided data.

        The purpose of this method is to update the SyncedCollection to match
        the data in the underlying resource.  The result of calling this method
        should be that ``self == data``. The reason that this method is
        necessary is that SyncedCollections can be nested, and nested
        collections must also be instances of SyncedCollection so that
        synchronization occurs even when nested structures are modified.
        Recreating the full nested structure every time data is reloaded from
        file is highly inefficient, so this method performs an in-place update
        that only changes entries that need to be changed.

        Parameters
        ----------
        data : collections.abc.Sequence, optional
            The data to be assigned to this list. If ``None``, the data is left
            unchanged (Default value = None).
        _validate : bool, optional
            If True, the data will not be validated (Default value = False).

        """
        if data is None:
            # If no data is passed, take no action.
            pass
        elif _sequence_resolver.get_type(data) == "SEQUENCE":
            with self._suspend_sync:
                # This loop is optimized based on common usage patterns:
                # insertion and removal at the end of a list. Inserting or
                # removing in the middle will result in extra conversion
                # operations for all subsequent items. In the worst case,
                # inserting at the beginning will require reconverting all
                # elements of the data.
                for i in range(min(len(self), len(data))):
                    if data[i] == self._data[i]:
                        continue
                    if _sc_resolver.get_type(self._data[i]) == "SYNCEDCOLLECTION":
                        try:
                            self._data[i]._update(data[i])
                            continue
                        except ValueError:
                            pass
                    if not _validate:
                        self._validate(data[i])
                    self._data[i] = self._from_base(data[i], parent=self)

                if len(self._data) > len(data):
                    self._data = self._data[: len(data)]
                else:
                    new_data = data[len(self) :]
                    if not _validate:
                        self._validate(new_data)
                    self.extend(new_data)
        else:
            raise ValueError(
                "Unsupported type: {}. The data must be a non-string sequence or None.".format(
                    type(data)
                )
            )

    def reset(self, data):
        """Update the instance with new data.

        Parameters
        ----------
        data : non-string Sequence
            Data to update the instance.

        Raises
        ------
        ValueError
            If the data is not a non-string sequence.

        """
        data = _convert_numpy(data)
        if _sequence_resolver.get_type(data) == "SEQUENCE":
            self._update(data)
            with self._thread_lock:
                self._save()
        else:
            raise ValueError(
                "Unsupported type: {}. The data must be a non-string sequence or None.".format(
                    type(data)
                )
            )

    def __setitem__(self, key, value):
        self._validate(value)
        with self._load_and_save, self._suspend_sync:
            self._data[key] = self._from_base(data=value, parent=self)

    def __reversed__(self):
        self._load()
        return reversed(self._data)

    def __iadd__(self, iterable):
        # Convert input to a list so that iterators work as well as iterables.
        iterable_data = list(iterable)
        self._validate(iterable_data)
        with self._load_and_save, self._suspend_sync:
            self._data += [
                self._from_base(data=value, parent=self) for value in iterable_data
            ]
        return self

    def insert(self, index, item):  # noqa: D102
        self._validate(item)
        with self._load_and_save, self._suspend_sync:
            self._data.insert(index, self._from_base(data=item, parent=self))

    def append(self, item):  # noqa: D102
        self._validate(item)
        with self._load_and_save, self._suspend_sync:
            self._data.append(self._from_base(data=item, parent=self))

    def extend(self, iterable):  # noqa: D102
        # Convert iterable to a list to ensure generators are exhausted only once
        iterable_data = list(iterable)
        self._validate(iterable_data)
        with self._load_and_save, self._suspend_sync:
            self._data.extend(
                [self._from_base(data=value, parent=self) for value in iterable_data]
            )

    def remove(self, value):  # noqa: D102
        with self._load_and_save, self._suspend_sync:
            self._data.remove(self._from_base(data=value, parent=self))

    def clear(self):  # noqa: D102
        self._data = []
        with self._thread_lock:
            self._save()

    def __lt__(self, other):
        if isinstance(other, type(self)):
            return self() < other()
        else:
            return self() > other

    def __le__(self, other):
        if isinstance(other, type(self)):
            return self() <= other()
        else:
            return self() <= other

    def __gt__(self, other):
        if isinstance(other, type(self)):
            return self() > other()
        else:
            return self() > other

    def __ge__(self, other):
        if isinstance(other, type(self)):
            return self() >= other()
        else:
            return self() >= other



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Define a framework for synchronized objects implementing the Collection interface.

Synchronization of standard Python data structures with a persistent data store is
important for a number of applications. While tools like `h5py` and `zarr` offer
dict-like interfaces to underlying files, these APIs serve to provide a familiar
wrapper around access patterns specific to these backends. Moreover, these formats
are primarily geared towards the provision of high-performance storage for large
array-like data. Storage of simpler data types, while possible, is generally
more difficult and requires additional work from the user.

Synced collections fills this gap, introducing a new abstract base class that extends
`collections.abc.Collection` to add transparent synchronization protocols. The package
implements its own versions of standard data structures like dicts and lists, and
it offers support for storing these data structures into various data formats. The
synchronization mechanism is completely transparent to the user; for example, a
`JSONDict` initialized pointing to a particular file can be modified like a normal
dict, and all changes will be automatically persisted to a JSON file.
"""

from .data_types import SyncedCollection, SyncedDict, SyncedList
from .version import __version__

__all__ = ["__version__", "SyncedCollection", "SyncedDict", "SyncedList"]



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Define utilities for handling NumPy arrays."""

import warnings

try:
    import numpy

    NUMPY = True
    _numpy_cache_blocklist = (numpy.ndarray,)
except ImportError:
    NUMPY = False
    _numpy_cache_blocklist = None  # type: ignore


class NumpyConversionWarning(UserWarning):
    """Warning raised when NumPy data is converted."""


NUMPY_CONVERSION_WARNING = (
    "Any numpy types provided will be transparently converted to the "
    "closest base Python equivalents."
)


def _convert_numpy(data):
    """Convert numpy data types to the corresponding base data types.

    0d numpy arrays and numpy scalars are converted to their corresponding
    primitive types, while other numpy arrays are converted to lists. If ``data``
    is not a numpy data type, this function is a no-op.
    """
    # Initially performing one isinstance check is faster since most of the
    # time the inputs are not numpy arrays, preventing additional function
    # calls inside the if statement.
    if NUMPY and isinstance(data, (numpy.ndarray, numpy.number, numpy.bool_)):
        if isinstance(data, numpy.ndarray):
            # tolist will return a scalar for 0d arrays, so there's no need to
            # special-case that check. 1-element 1d arrays should remain
            # arrays, i.e. np.array([1]) should become [1], not 1.
            warnings.warn(NUMPY_CONVERSION_WARNING, NumpyConversionWarning)
            return data.tolist()
        else:
            warnings.warn(NUMPY_CONVERSION_WARNING, NumpyConversionWarning)
            return data.item()
    return data


def _is_atleast_1d_numpy_array(data):
    """Check if an object is a nonscalar numpy array.

    The synced collections framework must differentiate 0d numpy arrays from
    other arrays because they are mapped to scalars while >0d arrays are mapped
    to (synced) lists.

    Returns
    -------
    bool
        Whether or not the input is a numpy array with at least 1 dimension.
    """
    return NUMPY and isinstance(data, numpy.ndarray) and data.ndim > 0


def _is_numpy_scalar(data):
    """Check if an object is a numpy scalar.

    This function is designed for use in situations where _convert_numpy has
    already been applied (if necessary), so 0d arrays are not considered
    scalars here.

    Returns
    -------
    bool
        Whether or not the input is a numpy scalar type.
    """
    return NUMPY and (
        (isinstance(data, (numpy.number, numpy.bool_)))
        or (isinstance(data, numpy.ndarray) and data.ndim == 0)
    )


def _is_complex(data):
    """Check if an object is complex.

    This function works for both numpy raw Python data types.

    Returns
    -------
    bool
        Whether or not the input is a complex number.
    """
    return (NUMPY and numpy.iscomplex(data).any()) or (isinstance(data, complex))



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Define supported backends.

No backends are imported by default. Users should import desired backends as
needed.
"""
from typing import List as _List

__all__: _List[str] = []



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import json
import os
from collections.abc import Collection, MutableSequence

import pytest

from synced_collections import SyncedList
from synced_collections.backends.collection_json import JSONDict
from synced_collections.numpy_utils import NumpyConversionWarning
from synced_collections.utils import AbstractTypeResolver, SyncedCollectionJSONEncoder

try:
    import numpy

    NUMPY = True
except ImportError:
    NUMPY = False


def test_type_resolver():
    resolver = AbstractTypeResolver(
        {
            "dict": lambda obj: isinstance(obj, dict),
            "tuple": lambda obj: isinstance(obj, tuple),
            "str": lambda obj: isinstance(obj, str),
            "mutablesequence": lambda obj: isinstance(obj, MutableSequence),
            "collection": lambda obj: isinstance(obj, Collection),
            "set": lambda obj: isinstance(obj, set),
        }
    )

    assert resolver.get_type({}) == "dict"
    assert resolver.get_type((0, 1)) == "tuple"
    assert resolver.get_type("abc") == "str"
    assert resolver.get_type([]) == "mutablesequence"

    # Make sure that order matters; collection should be found before list.
    assert resolver.get_type(set()) == "collection"


def test_json_encoder(tmpdir):
    # Raw dictionaries should be encoded transparently.
    data = {"foo": 1, "bar": 2, "baz": 3}
    json_str_data = '{"foo": 1, "bar": 2, "baz": 3}'
    assert json.dumps(data) == json_str_data
    assert json.dumps(data, cls=SyncedCollectionJSONEncoder) == json_str_data
    assert json.dumps(data, cls=SyncedCollectionJSONEncoder) == json.dumps(data)

    fn = os.path.join(tmpdir, "test_json_encoding.json")
    synced_data = JSONDict(fn)
    synced_data.update(data)
    with pytest.raises(TypeError):
        json.dumps(synced_data)
    assert json.dumps(synced_data, cls=SyncedCollectionJSONEncoder) == json_str_data

    if NUMPY:
        # Test both scalar and array numpy types since they could have
        # different problems.
        array = numpy.array(3)
        with pytest.warns(NumpyConversionWarning):
            synced_data["foo"] = array
        assert isinstance(synced_data["foo"], int)

        array = numpy.random.rand(3)
        with pytest.warns(NumpyConversionWarning):
            synced_data["foo"] = array
        assert isinstance(synced_data["foo"], SyncedList)
        assert (
            json.loads(json.dumps(synced_data, cls=SyncedCollectionJSONEncoder))
            == synced_data()
        )



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implements the :class:`AttrDict`.

This simple mixin class implements overloads for __setattr__, __getattr__, and
__delattr__.  While we do not want to offer this API generally for all
SyncedDict objects, some applications may want to add this feature, so this
simple mixin can be combined via inheritance without causing much difficulty.
"""

from typing import FrozenSet


class AttrDict:
    """A class that redirects attribute access methods to ``__getitem__``.

    Although this class is called an :class:`AttrDict`, it does not directly
    inherit from any dict-like class or offer any relevant APIs. Its only purpose
    is to be used as a mixin with other dict-like classes to add attribute-based
    access to dictionary contents.

    Subclasses that inherit from this class must define the ``_PROTECTED_KEYS``
    class variable, which indicates known attributes of the object. This indication
    is necessary because otherwise accessing ``obj.data`` is ambiguous as to
    whether it is a reference to a special ``data`` attribute or whether it is
    equivalent to ``obj['data']``. Without this variable, a user could mask
    internal variables inaccessible via normal attribute access by adding dictionary
    keys with the same name.

    Examples
    --------
    >>> assert dictionary['foo'] == dictionary.foo

    """

    _PROTECTED_KEYS: FrozenSet[str] = frozenset()

    def __getattr__(self, name):
        if name.startswith("__"):
            raise AttributeError(f"{type(self)} has no attribute '{name}'")
        try:
            return self.__getitem__(name)
        except KeyError as e:
            raise AttributeError(e)

    def __setattr__(self, key, value):
        # This logic assumes that __setitem__ will not be called until after
        # the object has been fully instantiated. We may want to add a try
        # except in the else clause in case someone subclasses these and tries
        # to use d['foo'] inside a constructor prior to _data being defined.
        # The order of these checks assumes that setting protected keys will be
        # much more common than setting dunder attributes.
        if key in self._PROTECTED_KEYS or key.startswith("__"):
            super().__setattr__(key, value)
        else:
            self.__setitem__(key, value)

    def __delattr__(self, key):
        # The order of these checks assumes that deleting protected keys will be
        # much more common than deleting dunder attributes.
        if key in self._PROTECTED_KEYS or key.startswith("__"):
            super().__delattr__(key)
        else:
            self.__delitem__(key)



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Errors raised by synced collections."""


class BufferException(RuntimeError):
    """Raised when any exception related to buffering occurs."""


class BufferedError(BufferException):
    """Raised when an error occured while flushing one or more buffered files.

    Attribute
    ---------
    files : dict
        A dictionary of names that caused issues during the flush operation,
        mapped to a possible reason for the issue or None in case that it
        cannot be determined.
    """

    def __init__(self, files):
        self.files = files

    def __str__(self):
        return f"{type(self).__name__}({self.files})"


class MetadataError(BufferException):
    """Raised when metadata check fails.

    The contents of this file in the buffer can be accessed via the
    `buffer_contents` attribute of this exception.
    """

    def __init__(self, filename, contents):
        self.filename = filename
        self.buffer_contents = contents

    def __str__(self):
        return f"{self.filename} appears to have been externally modified."


class KeyTypeError(TypeError):
    """Raised when a user uses a key of invalid type."""


class InvalidKeyError(ValueError):
    """Raised when a user uses a non-conforming key."""



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implements a JSON :class:`~.SyncedCollection` backend."""

import errno
import json
import os
import sys
import uuid
from collections.abc import Mapping, Sequence
from typing import Callable, FrozenSet
from typing import Sequence as Sequence_t

from .. import SyncedCollection, SyncedDict, SyncedList
from ..buffers.memory_buffered_collection import SharedMemoryFileBufferedCollection
from ..buffers.serialized_file_buffered_collection import (
    SerializedFileBufferedCollection,
)
from ..data_types.attr_dict import AttrDict
from ..errors import InvalidKeyError, KeyTypeError
from ..numpy_utils import (
    _is_atleast_1d_numpy_array,
    _is_complex,
    _is_numpy_scalar,
    _numpy_cache_blocklist,
)
from ..utils import AbstractTypeResolver, SyncedCollectionJSONEncoder
from ..validators import json_format_validator, no_dot_in_key, require_string_key

ON_WINDOWS = sys.platform.startswith("win32") or sys.platform.startswith("cygwin")


"""
There are many classes defined in this file. Most of the definitions are
trivial since logic is largely inherited, but the large number of classes
and the extensive docstrings can be intimidating and make the source
difficult to parse. Section headers like these are used to organize the
code to reduce this barrier.
"""

_json_attr_dict_validator_type_resolver = AbstractTypeResolver(
    {
        # We identify >0d numpy arrays as sequences for validation purposes.
        "SEQUENCE": lambda obj: (isinstance(obj, Sequence) and not isinstance(obj, str))
        or _is_atleast_1d_numpy_array(obj),
        "NUMPY": lambda obj: _is_numpy_scalar(obj),
        "BASE": lambda obj: isinstance(obj, (str, int, float, bool, type(None))),
        "MAPPING": lambda obj: isinstance(obj, Mapping),
    },
    cache_blocklist=_numpy_cache_blocklist,
)


def json_attr_dict_validator(data):
    """Validate data for JSONAttrDict.

    This validator combines the logic from the following validators into one to
    make validation more efficient:

    This validator combines the following logic:
        - JSON format validation
        - Ensuring no dots are present in string keys

    Parameters
    ----------
    data
        Data to validate.

    Raises
    ------
    KeyTypeError
        If key data type is not supported.
    TypeError
        If the data type of ``data`` is not supported.

    """
    switch_type = _json_attr_dict_validator_type_resolver.get_type(data)

    if switch_type == "BASE":
        return
    elif switch_type == "MAPPING":
        # Explicitly call `list(keys)` to get a fixed list of keys to avoid
        # running into issues with iterating over a DictKeys view while
        # modifying the dict at the same time. Inside the loop, we:
        #   1) validate the key, converting to string if necessary
        #   2) pop and validate the value
        #   3) reassign the value to the (possibly converted) key
        for key in list(data):
            json_attr_dict_validator(data[key])
            if isinstance(key, str):
                if "." in key:
                    raise InvalidKeyError(
                        f"Mapping keys may not contain dots ('.'): {key}."
                    )
            else:
                raise KeyTypeError(
                    f"Mapping keys must be str, not {type(key).__name__}."
                )
    elif switch_type == "SEQUENCE":
        for value in data:
            json_attr_dict_validator(value)
    elif switch_type == "NUMPY":
        if _is_numpy_scalar(data.item()):
            raise TypeError("NumPy extended precision types are not JSON serializable.")
        elif _is_complex(data):
            raise TypeError("Complex numbers are not JSON serializable.")
    else:
        raise TypeError(
            f"Object of type {type(data).__name__} is not JSON serializable."
        )


"""
Here we define the main JSONCollection class that encapsulates most of the
logic for reading from and writing to JSON files. The remaining classes in
this file inherit from these classes to add features like buffering or
attribute-based dictionary access, each with a different backend name for
correct resolution of nested SyncedCollection types.
"""


class JSONCollection(SyncedCollection):
    r"""A :class:`~.SyncedCollection` that synchronizes with a JSON file.

    This collection implements synchronization by reading and writing the associated
    JSON file in its entirety for every read/write operation. This backend is a good
    choice for maximum accessibility and transparency since all data is immediately
    accessible in the form of a text file with no additional tooling, but is
    likely a poor choice for high performance applications.

    **Thread safety**

    The :class:`JSONCollection` is thread-safe on Unix-like systems (not
    Windows, see the Warnings section). To make these collections safe, the
    ``write_concern`` flag is ignored in multithreaded execution, and the write
    is **always** performed via a write to temporary file followed by a
    replacement of the original file. The file replacement operation uses
    :func:`os.replace`, which is guaranteed to be atomic by the Python
    standard.

    Parameters
    ----------
    filename : str
        The filename of the associated JSON file on disk.
    write_concern : bool, optional
        Ensure file consistency by writing changes back to a temporary file
        first, before replacing the original file (Default value = False).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    This class is *not* thread safe on Windows. It relies on ``os.replace`` for
    atomic file writes, and that method can fail in multithreaded situations if
    open handles exist to the destination file within the same process on a
    different thread. See https://bugs.python.org/issue46003 for more
    information.
    """

    _backend = __name__  # type: ignore
    _supports_threading = not ON_WINDOWS
    _validators: Sequence_t[Callable] = (require_string_key, json_format_validator)

    def __init__(self, filename=None, write_concern=False, *args, **kwargs):
        # The `_filename` attribute _must_ be defined prior to calling the
        # superclass constructors because the filename defines the `_lock_id`
        # used to uniquely identify thread locks for this collection.
        self._filename = filename
        super().__init__(*args, **kwargs)
        self._write_concern = write_concern

    def _load_from_resource(self):
        """Load the data from a JSON file.

        Returns
        -------
        Collection or None
            An equivalent unsynced collection satisfying :meth:`is_base_type` that
            contains the data in the JSON file. Will return None if the file does
            not exist.

        """
        try:
            with open(self._filename, "rb") as file:
                blob = file.read()
                return json.loads(blob)
        except OSError as error:
            if error.errno == errno.ENOENT:
                return None
            else:
                raise

    def _save_to_resource(self):
        """Write the data to JSON file."""
        # Serialize data
        blob = json.dumps(self, cls=SyncedCollectionJSONEncoder).encode()
        # When write_concern flag is set, we write the data into dummy file and then
        # replace that file with original file. We also enable this mode
        # irrespective of the write_concern flag if we're running in
        # multithreaded mode.
        if self._write_concern or type(self)._threading_support_is_active:
            dirname, filename = os.path.split(self._filename)
            fn_tmp = os.path.join(dirname, f"._{uuid.uuid4()}_{filename}")
            with open(fn_tmp, "wb") as tmpfile:
                tmpfile.write(blob)
            os.replace(fn_tmp, self._filename)
        else:
            with open(self._filename, "wb") as file:
                file.write(blob)

    @property
    def filename(self):
        """str: Get or set the name of the associated JSON file on disk."""
        return self._filename

    @filename.setter
    def filename(self, value):
        # When setting the filename we must also remap the locks.
        with self._thread_lock:
            if type(self)._threading_support_is_active:
                old_lock_id = self._lock_id

            self._filename = value

            if type(self)._threading_support_is_active:
                type(self)._locks[self._lock_id] = type(self)._locks.pop(old_lock_id)

    @property
    def _lock_id(self):
        return self._filename


# These are the common protected keys used by all JSONDict types.
_JSONDICT_PROTECTED_KEYS = frozenset(
    (
        # These are all protected keys that are inherited from data type classes.
        "_data",
        "_name",
        "_suspend_sync_",
        "_load",
        "_sync",
        "_root",
        "_validators",
        "_all_validators",
        "_load_and_save",
        "_suspend_sync",
        "_supports_threading",
        "_LoadSaveType",
        "registry",
        # These keys are specific to the JSON backend.
        "_filename",
        "filename",
        "_write_concern",
    )
)


class JSONDict(JSONCollection, SyncedDict):
    r"""A dict-like data structure that synchronizes with a persistent JSON file.

    Examples
    --------
    >>> doc = JSONDict('data.json', write_concern=True)
    >>> doc['foo'] = "bar"
    >>> assert doc['foo'] == "bar"
    >>> assert 'foo' in doc
    >>> del doc['foo']
    >>> doc['foo'] = dict(bar=True)
    >>> doc
    {'foo': {'bar': True}}

    Parameters
    ----------
    filename : str, optional
        The filename of the associated JSON file on disk (Default value = None).
    write_concern : bool, optional
        Ensure file consistency by writing changes back to a temporary file
        first, before replacing the original file (Default value = False).
    data : :class:`collections.abc.Mapping`, optional
        The initial data passed to :class:`JSONDict`. If ``None``, defaults to
        ``{}`` (Default value = None).
    parent : JSONCollection, optional
        A parent instance of :class:`JSONCollection` or ``None``. If ``None``,
        the collection owns its own data (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    While the :class:`JSONDict` object behaves like a :class:`dict`, there are important
    distinctions to remember. In particular, because operations are reflected
    as changes to an underlying file, copying (even deep copying) a :class:`JSONDict`
    instance may exhibit unexpected behavior. If a true copy is required, you
    should use the call operator to get a dictionary representation, and if
    necessary construct a new :class:`JSONDict` instance.

    """

    _PROTECTED_KEYS: FrozenSet[str] = _JSONDICT_PROTECTED_KEYS

    def __init__(
        self,
        filename=None,
        write_concern=False,
        data=None,
        parent=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            filename=filename,
            write_concern=write_concern,
            data=data,
            parent=parent,
            *args,
            **kwargs,
        )


class JSONList(JSONCollection, SyncedList):
    r"""A list-like data structure that synchronizes with a persistent JSON file.

    Only non-string sequences are supported by this class.

    Examples
    --------
    >>> synced_list = JSONList('data.json', write_concern=True)
    >>> synced_list.append("bar")
    >>> assert synced_list[0] == "bar"
    >>> assert len(synced_list) == 1
    >>> del synced_list[0]

    Parameters
    ----------
    filename : str, optional
        The filename of the associated JSON file on disk (Default value = None).
    write_concern : bool, optional
        Ensure file consistency by writing changes back to a temporary file
        first, before replacing the original file (Default value = None).
    data : non-str :class:`collections.abc.Sequence`, optional
        The initial data passed to :class:`JSONList`. If ``None``, defaults to
        ``[]`` (Default value = None).
    parent : JSONCollection, optional
        A parent instance of :class:`JSONCollection` or ``None``. If ``None``,
        the collection owns its own data (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    While the :class:`JSONList` object behaves like a :class:`list`, there are important
    distinctions to remember. In particular, because operations are reflected
    as changes to an underlying file, copying (even deep copying) a :class:`JSONList`
    instance may exhibit unexpected behavior. If a true copy is required, you
    should use the call operator to get a dictionary representation, and if
    necessary construct a new :class:`JSONList` instance.

    """

    def __init__(
        self,
        filename=None,
        write_concern=False,
        data=None,
        parent=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            filename=filename,
            write_concern=write_concern,
            data=data,
            parent=parent,
            *args,
            **kwargs,
        )


"""
Here we define the BufferedJSONCollection class and its data type
subclasses, which augment the JSONCollection with a serialized in-memory
buffer for improved performance.
"""


class BufferedJSONCollection(SerializedFileBufferedCollection, JSONCollection):
    """A :class:`JSONCollection` that supports I/O buffering.

    This class implements the buffer protocol defined by
    :class:`~.BufferedCollection`. The concrete implementation of buffering
    behavior is defined by the :class:`~.SerializedFileBufferedCollection`.
    """

    _backend = __name__ + ".buffered"  # type: ignore


# These are the keys common to buffer backends.
_BUFFERED_PROTECTED_KEYS = frozenset(
    (
        "buffered",
        "_is_buffered",
        "_buffer_lock",
        "_buffer_context",
        "_buffered_collections",
    )
)


class BufferedJSONDict(BufferedJSONCollection, SyncedDict):
    """A buffered :class:`JSONDict`."""

    _PROTECTED_KEYS: FrozenSet[str] = (
        _JSONDICT_PROTECTED_KEYS | _BUFFERED_PROTECTED_KEYS
    )

    def __init__(
        self,
        filename=None,
        write_concern=False,
        data=None,
        parent=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            filename=filename,
            write_concern=write_concern,
            data=data,
            parent=parent,
            *args,
            **kwargs,
        )


class BufferedJSONList(BufferedJSONCollection, SyncedList):
    """A buffered :class:`JSONList`."""

    def __init__(
        self,
        filename=None,
        write_concern=False,
        data=None,
        parent=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            filename=filename,
            write_concern=write_concern,
            data=data,
            parent=parent,
            *args,
            **kwargs,
        )


"""
Here we define the MemoryBufferedJSONCollection class and its data type
subclasses, which augment the JSONCollection with a serialized in-memory
buffer for improved performance.
"""


class MemoryBufferedJSONCollection(SharedMemoryFileBufferedCollection, JSONCollection):
    """A :class:`JSONCollection` that supports I/O buffering.

    This class implements the buffer protocol defined by :class:`~.BufferedCollection`.
    The concrete implementation of buffering behavior is defined by the
    :class:`~.SharedMemoryFileBufferedCollection`.
    """

    _backend = __name__ + ".memory_buffered"  # type: ignore


class MemoryBufferedJSONDict(MemoryBufferedJSONCollection, SyncedDict):
    """A buffered :class:`JSONDict`."""

    _PROTECTED_KEYS: FrozenSet[str] = (
        _JSONDICT_PROTECTED_KEYS | _BUFFERED_PROTECTED_KEYS
    )

    def __init__(
        self,
        filename=None,
        write_concern=False,
        data=None,
        parent=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            filename=filename,
            write_concern=write_concern,
            data=data,
            parent=parent,
            *args,
            **kwargs,
        )


class MemoryBufferedJSONList(MemoryBufferedJSONCollection, SyncedList):
    """A buffered :class:`JSONList`."""

    def __init__(
        self,
        filename=None,
        write_concern=False,
        data=None,
        parent=None,
        *args,
        **kwargs,
    ):
        super().__init__(
            filename=filename,
            write_concern=write_concern,
            data=data,
            parent=parent,
            *args,
            **kwargs,
        )


"""
Here we define various extensions of the above classes that add
attribute-based access to dictionaries. Although list behavior is not
modified in any way by these, they still require separate classes with the
right backend so that nested classes are created appropriately.
"""


class JSONAttrDict(JSONDict, AttrDict):
    r"""A dict-like data structure that synchronizes with a persistent JSON file.

    Unlike :class:`JSONDict`, this class also supports attribute-based access to
    dictionary contents, e.g. ``doc.foo == doc['foo']``.

    Examples
    --------
    >>> doc = JSONAttrDict('data.json', write_concern=True)
    >>> doc['foo'] = "bar"
    >>> assert doc.foo == doc['foo'] == "bar"
    >>> assert 'foo' in doc
    >>> del doc['foo']
    >>> doc['foo'] = dict(bar=True)
    >>> doc
    {'foo': {'bar': True}}
    >>> doc.foo.bar = False
    >>> doc
    {'foo': {'bar': False}}

    Parameters
    ----------
    filename : str, optional
        The filename of the associated JSON file on disk (Default value = None).
    write_concern : bool, optional
        Ensure file consistency by writing changes back to a temporary file
        first, before replacing the original file (Default value = False).
    data : :class:`collections.abc.Mapping`, optional
        The initial data passed to :class:`JSONAttrDict`. If ``None``, defaults to
        ``{}`` (Default value = None).
    parent : JSONCollection, optional
        A parent instance of :class:`JSONCollection` or ``None``. If ``None``,
        the collection owns its own data (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    While the :class:`JSONAttrDict` object behaves like a :class:`dict`, there are important
    distinctions to remember. In particular, because operations are reflected
    as changes to an underlying file, copying (even deep copying) a :class:`JSONAttrDict`
    instance may exhibit unexpected behavior. If a true copy is required, you
    should use the call operator to get a dictionary representation, and if
    necessary construct a new :class:`JSONAttrDict` instance.

    """

    _backend = __name__ + ".attr"  # type: ignore
    # Define the validators in case subclasses want to inherit the correct
    # behavior, but define _all_validators for performance of this class.
    _validators = (no_dot_in_key,)
    _all_validators = (json_attr_dict_validator,)


class JSONAttrList(JSONList):
    """A :class:`JSONList` whose dict-like children will be of type :class:`JSONAttrDict`."""

    _backend = __name__ + ".attr"  # type: ignore


class BufferedJSONAttrDict(BufferedJSONDict, AttrDict):
    """A buffered :class:`JSONAttrDict`."""

    _backend = __name__ + ".buffered_attr"  # type: ignore
    # Define the validators in case subclasses want to inherit the correct
    # behavior, but define _all_validators for performance of this class.
    _validators = (no_dot_in_key,)
    _all_validators = (json_attr_dict_validator,)


class BufferedJSONAttrList(BufferedJSONList):
    """A :class:`BufferedJSONList` whose dict-like children will be of type :class:`BufferedJSONAttrDict`."""  # noqa: E501

    _backend = __name__ + ".buffered_attr"  # type: ignore


class MemoryBufferedJSONAttrDict(MemoryBufferedJSONDict, AttrDict):
    """A buffered :class:`JSONAttrDict`."""

    _backend = __name__ + ".memory_buffered_attr"  # type: ignore
    # Define the validators in case subclasses want to inherit the correct
    # behavior, but define _all_validators for performance of this class.
    _validators = (no_dot_in_key,)
    _all_validators = (json_attr_dict_validator,)


class MemoryBufferedJSONAttrList(MemoryBufferedJSONList):
    """A :class:`MemoryBufferedJSONList` whose dict-like children will be of type :class:`MemoryBufferedJSONAttrDict`."""  # noqa: E501

    _backend = __name__ + ".memory_buffered_attr"  # type: ignore



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""A standardized buffering implementation for file-based backends.

All file-based backends can use a similar buffering protocol. In particular,
integrity checks can be performed by checking for whether the file has been
modified since it was originally loaded into the buffer. However, various
specific components are abstract and must be implemented by child classes.
"""

import errno
import os
from abc import abstractmethod
from threading import RLock
from typing import Dict, Tuple, Union

from ..data_types.synced_collection import _LoadAndSave
from ..errors import BufferedError, MetadataError
from ..utils import _CounterFuncContext, _NullContext
from .buffered_collection import BufferedCollection


class _FileBufferedContext(_CounterFuncContext):
    """Extend the usual buffering context to support setting the buffer size.

    This context allows the buffer_backend method to temporarily set the buffer
    size within the scope of this context.
    """

    def __init__(self, cls):
        super().__init__(cls._flush_buffer)
        self._buffer_capacity = None
        self._original_buffer_capacitys = []
        self._cls = cls

    def __call__(self, buffer_capacity=None):
        self._buffer_capacity = buffer_capacity
        return self

    def __enter__(self):
        super().__enter__()
        if self._buffer_capacity is not None:
            self._original_buffer_capacitys.append(self._cls.get_buffer_capacity())
            self._cls.set_buffer_capacity(self._buffer_capacity)
        else:
            self._original_buffer_capacitys.append(None)
        self._buffer_capacity = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        original_buffer_capacity = self._original_buffer_capacitys.pop()
        if original_buffer_capacity is not None:
            self._cls.set_buffer_capacity(original_buffer_capacity)


class _BufferedLoadAndSave(_LoadAndSave):
    """Wrap base loading and saving with an extra thread lock.

    Writes to buffered collections will also modify the buffer, so they must
    acquire the buffer lock in addition to the default behavior.
    """

    def __enter__(self):
        self._collection._buffer_lock.__enter__()
        super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            super().__exit__(exc_type, exc_val, exc_tb)
        finally:
            self._collection._buffer_lock.__exit__(exc_type, exc_val, exc_tb)


class FileBufferedCollection(BufferedCollection):
    """A :class:`~.SyncedCollection` that can buffer file I/O.

    This class provides a standardized buffering protocol for all file-based
    backends. All file-based backends can use the same set of integrity checks
    prior to a buffer flush to ensure that no conflicting modifications are
    made. Specifically, they can check whether the file has been modified on
    disk since it was originally loaded to the buffer. This class provides the
    basic infrastructure for that and defines standard methods that can be used
    by all classes. Subclasses must define the appropriate storage mechanism.

    .. note::
        Important notes for developers:
            - This class should be inherited before any other collections. This
              requirement is due to the extensive use of multiple inheritance.
              Since this class is designed to be combined with other
              :class:`~.SyncedCollection` types without making those types aware
              of buffering behavior, it transparently hooks into the
              initialization process, but this is dependent on its constructor
              being called before those of other classes.
            - All subclasses must define a class level ``_BUFFER_CAPACITY``
              variable that is used to determine the maximum allowable buffer
              size.

    Parameters
    ----------
    filename : str, optional
        The filename of the associated JSON file on disk (Default value = None).

    Warnings
    --------
    Although it can be done safely, in general modifying two different collections
    pointing to the same underlying resource while both are in different buffering
    modes is unsupported and can lead to undefined behavior. This class makes a
    best effort at performing safe modifications, but it is possible to construct
    nested buffered contexts for different objects that can lead to an invalid
    buffer state, or even situations where there is no obvious indicator of what
    is the canonical source of truth. In general, if you need multiple objects
    pointing to the same resource, it is **strongly** recommeneded to work with
    both of them in identical buffering states at all times.

    """

    _LoadSaveType = _BufferedLoadAndSave

    def __init__(self, parent=None, filename=None, *args, **kwargs):
        super().__init__(parent=parent, filename=filename, *args, **kwargs)
        self._filename = filename

    @classmethod
    def __init_subclass__(cls):
        """Prepare subclasses."""
        super().__init_subclass__()
        cls._CURRENT_BUFFER_SIZE = 0

        # This dict is the actual data buffer, mapping filenames to their
        # cached data and metadata.
        cls._buffer: Dict[str, Dict[str, Union[bytes, str, Tuple[int, float]]]] = {}

        # Buffered contexts may be nested, and when leaving a buffered context
        # we only want to flush collections that are no longer buffered. To
        # accomplish this, we maintain a list of buffered collections so that
        # we can perform per-instance flushes that account for their current
        # buffering state.
        cls._buffered_collections: Dict[int, BufferedCollection] = {}

        cls._buffer_context = _FileBufferedContext(cls)

    @classmethod
    def enable_multithreading(cls):
        """Enable safety checks and thread locks required for thread safety.

        This method adds managed buffer-related thread safety in addition to
        what the parent method does.

        """
        super().enable_multithreading()
        cls._BUFFER_LOCK = RLock()

    @classmethod
    def disable_multithreading(cls):
        """Disable all safety checks and thread locks required for thread safety.

        This method adds managed buffer-related thread safety in addition to
        what the parent method does.

        """
        super().disable_multithreading()
        cls._BUFFER_LOCK = _NullContext()

    @property
    def _buffer_lock(self):
        """Acquire the buffer lock."""
        return type(self)._BUFFER_LOCK

    def _get_file_metadata(self):
        """Return metadata of file.

        Returns
        -------
        Tuple[int, float] or None
            The size and last modification time of the associated file. If the
            file does not exist, returns :code:`None`.

        """
        try:
            metadata = os.stat(self._filename)
            return metadata.st_size, metadata.st_mtime_ns
        except OSError as error:
            if error.errno != errno.ENOENT:
                raise
            # A return value of None indicates that the file does not
            # exist. Since any non-``None`` value will return `False` when
            # compared to ``None``, returning ``None`` provides a
            # reasonable value to compare against for metadata-based
            # validation checks.
            return None

    @classmethod
    def get_buffer_capacity(cls):
        """Get the current buffer capacity.

        Returns
        -------
        int
            The amount of data that can be stored before a flush is triggered
            in the appropriate units for a particular buffering implementation.

        """
        return cls._BUFFER_CAPACITY

    @classmethod
    def set_buffer_capacity(cls, new_capacity):
        """Update the buffer capacity.

        Parameters
        ----------
        new_capacity : int
            The new capacity of the buffer in the appropriate units for a particular
            buffering implementation.

        """
        cls._BUFFER_CAPACITY = new_capacity
        with cls._BUFFER_LOCK:
            if new_capacity < cls._CURRENT_BUFFER_SIZE:
                cls._flush_buffer(force=True)

    @classmethod
    def get_current_buffer_size(cls):
        """Get the total amount of data currently stored in the buffer.

        Returns
        -------
        int
            The size of all data contained in the buffer in the appropriate
            units for a particular buffering implementation.

        """
        return cls._CURRENT_BUFFER_SIZE

    def _load_from_buffer(self):
        """Read data from buffer.

        See :meth:`~._initialize_data_in_buffer` for details on the data stored
        in the buffer and the integrity checks performed.

        Returns
        -------
        collections.abc.Collection
            A collection of the same base type as the :class:`~.SyncedCollection` this
            method is called for, corresponding to data loaded from the
            underlying file.

        """
        with self._buffer_lock:
            if self._filename not in type(self)._buffer:
                # The first time this method is called, if nothing is in the buffer
                # for this file then we cannot guarantee that the _data attribute
                # is valid either since the resource could have been modified
                # between when _data was last updated and when this load is being
                # called. As a result, we have to load from the resource here to be
                # safe.
                data = self._load_from_resource()
                with self._thread_lock, self._suspend_sync:
                    self._update(data)
                self._initialize_data_in_buffer()

        # This storage can be safely updated every time on every thread.
        type(self)._buffered_collections[id(self)] = self

    @abstractmethod
    def _initialize_data_in_buffer(self):
        """Create the initial entry for the data in the cache.

        This method should be called the first time that a collection's data is
        accessed in buffered mode. This method stores the encoded data in the
        cache, along with the metadata of the underlying file and any other
        information that may be used for validation later. This information
        depends on the implementation of the buffer in subclasses.
        """
        pass

    @classmethod
    def _flush_buffer(cls, force=False, retain_in_force=False):
        """Flush the data in the file buffer.

        Parameters
        ----------
        force : bool
            If True, force a flush even in buffered mode (defaults to False). This
            parameter is used when the buffer is filled to capacity.
        retain_in_force : bool
            If True, when forcing a flush a collection is retained in the buffer.
            This feature is useful if only some subset of the buffer's contents
            are relevant to size restrictions. For intance, since only modified
            items will have to be written back out to disk, a buffer protocol may
            not care to count unmodified collections towards the total.

        Returns
        -------
        dict
            Mapping of filename and errors occured during flushing data.

        Raises
        ------
        BufferedError
            If there are any issues with flushing the data.

        """
        issues = {}

        # We need to use the list of buffered objects rather than directly
        # looping over the local cache so that each collection can
        # independently decide whether or not to flush based on whether it's
        # still buffered (if buffered contexts are nested).
        remaining_collections = {}
        while True:
            # This is the only part that needs to be locked; once items are
            # removed from the buffer they can be safely handled on separate
            # threads.
            with cls._BUFFER_LOCK:
                try:
                    (
                        col_id,
                        collection,
                    ) = cls._buffered_collections.popitem()
                except KeyError:
                    break

            if collection._is_buffered and not force:
                # If force is false, then the only way for the collection to
                # still be buffered is if there are nested buffered contexts.
                # In that case, flush_buffer was called due to the exit of an
                # inner buffered context, and we shouldn't do anything with
                # this object, so we just put it back in the list *and* skip
                # the flush.
                remaining_collections[col_id] = collection
                continue
            elif force and retain_in_force:
                # If force is true, the collection must still be buffered.
                # In that case, the retain_in_force parameter controls whether
                # we we want to put it back in the remaining_collections list
                # after flushing any writes.
                remaining_collections[col_id] = collection

            try:
                collection._flush(force=force)
            except (OSError, MetadataError) as err:
                issues[collection._filename] = err
        if not issues:
            cls._buffered_collections = remaining_collections
        else:
            raise BufferedError(issues)

    @classmethod
    def buffer_backend(cls, buffer_capacity=None):
        """Enter context to buffer all operations for this backend.

        Parameters
        ----------
        buffer_capacity : int
            The capacity of the buffer to use within this context (resets after
            the context is exited).
        """
        return cls._buffer_context(buffer_capacity)



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Validators for SyncedCollection API.

A validator is any callable that raises Exceptions when called with invalid data.
Validators should act recursively for nested data structures and should not
return any values, only raise errors. This module implements built-in validators,
but client code is free to implement and add additioal validators to collection
types as needed.
"""

from collections.abc import Mapping, Sequence

from .errors import InvalidKeyError, KeyTypeError
from .numpy_utils import (
    _is_atleast_1d_numpy_array,
    _is_complex,
    _is_numpy_scalar,
    _numpy_cache_blocklist,
)
from .utils import AbstractTypeResolver

_no_dot_in_key_type_resolver = AbstractTypeResolver(
    {
        "MAPPING": lambda obj: isinstance(obj, Mapping),
        "SEQUENCE": lambda obj: isinstance(obj, Sequence) and not isinstance(obj, str),
    }
)


def no_dot_in_key(data):
    """Raise an exception if there is a dot (``.``) in a mapping's key.

    Parameters
    ----------
    data
        Data to validate.

    Raises
    ------
    KeyTypeError
        If key data type is not supported.
    InvalidKeyError
        If the key contains invalid characters or is otherwise malformed.

    """
    switch_type = _no_dot_in_key_type_resolver.get_type(data)

    if switch_type == "MAPPING":
        for key, value in data.items():
            if isinstance(key, str):
                if "." in key:
                    raise InvalidKeyError(
                        f"Mapping keys may not contain dots ('.'): {key}"
                    )
            else:
                raise KeyTypeError(
                    f"Mapping keys must be str, not {type(key).__name__}"
                )
            no_dot_in_key(value)
    elif switch_type == "SEQUENCE":
        for value in data:
            no_dot_in_key(value)


def require_string_key(data):
    """Raise an exception if key in a mapping is not a string.

    Almost all supported backends require string keys.

    Parameters
    ----------
    data
        Data to validate.

    Raises
    ------
    KeyTypeError
        If key type is not a string.

    """
    # Reuse the type resolver here since it has the same groupings.
    switch_type = _no_dot_in_key_type_resolver.get_type(data)

    if switch_type == "MAPPING":
        for key, value in data.items():
            if not isinstance(key, str):
                raise KeyTypeError(
                    f"Mapping keys must be str, not {type(key).__name__}"
                )
            require_string_key(value)
    elif switch_type == "NON_STR_SEQUENCE":
        for value in data:
            require_string_key(value)


_json_format_validator_type_resolver = AbstractTypeResolver(
    {
        # We identify >0d numpy arrays as sequences for validation purposes.
        "SEQUENCE": lambda obj: (isinstance(obj, Sequence) and not isinstance(obj, str))
        or _is_atleast_1d_numpy_array(obj),
        "NUMPY": lambda obj: _is_numpy_scalar(obj),
        "BASE": lambda obj: isinstance(obj, (str, int, float, bool, type(None))),
        "MAPPING": lambda obj: isinstance(obj, Mapping),
    },
    cache_blocklist=_numpy_cache_blocklist,
)


def json_format_validator(data):
    """Validate input data can be serialized to JSON.

    Parameters
    ----------
    data
        Data to validate.

    Raises
    ------
    KeyTypeError
        If key data type is not supported.
    TypeError
        If the data type of ``data`` is not supported.

    """
    switch_type = _json_format_validator_type_resolver.get_type(data)

    if switch_type == "BASE":
        return
    elif switch_type == "MAPPING":
        for key, value in data.items():
            if not isinstance(key, str):
                raise KeyTypeError(f"Keys must be str, not {type(key).__name__}")
            json_format_validator(value)
    elif switch_type == "SEQUENCE":
        for value in data:
            json_format_validator(value)
    elif switch_type == "NUMPY":
        if _is_numpy_scalar(data.item()):
            raise TypeError("NumPy extended precision types are not JSON serializable.")
        elif _is_complex(data):
            raise TypeError("Complex numbers are not JSON serializable.")
    else:
        raise TypeError(
            f"Object of type {type(data).__name__} is not JSON serializable"
        )



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import json
import os

import pytest
from attr_dict_test import AttrDictTest, AttrListTest
from synced_collection_test import SyncedDictTest, SyncedListTest

from synced_collections.backends.collection_json import (
    ON_WINDOWS,
    JSONAttrDict,
    JSONAttrList,
    JSONDict,
    JSONList,
)


class JSONCollectionTest:
    _write_concern = False
    _fn = "test.json"

    def store(self, synced_collection, data):
        with open(synced_collection.filename, "wb") as f:
            f.write(json.dumps(data).encode())

    @pytest.fixture
    def synced_collection(self, tmpdir):
        yield self._collection_type(
            filename=os.path.join(tmpdir, self._fn),
            write_concern=self._write_concern,
        )

    @pytest.fixture
    def synced_collection_positional(self, tmpdir):
        """Fixture that initializes the object using positional arguments."""
        yield self._collection_type(
            os.path.join(tmpdir, "test2.json"), self._write_concern
        )

    def test_filename(self, synced_collection):
        assert os.path.basename(synced_collection.filename) == self._fn

    @pytest.mark.skipif(
        ON_WINDOWS,
        reason=(
            "The JSONCollection cannot be safely used in multithreaded settings "
            "on Windows due to https://bugs.python.org/issue46003."
        ),
    )
    def test_multithreaded(self, synced_collection):
        return super().test_multithreaded(synced_collection)


class TestJSONDict(JSONCollectionTest, SyncedDictTest):
    _collection_type = JSONDict


class TestJSONList(JSONCollectionTest, SyncedListTest):
    _collection_type = JSONList


class TestJSONDictWriteConcern(TestJSONDict):
    _write_concern = True


class TestJSONListWriteConcern(TestJSONList):
    _write_concern = True


class TestJSONAttrDict(TestJSONDict, AttrDictTest):
    _collection_type = JSONAttrDict


class TestJSONAttrList(TestJSONList, AttrListTest):
    _collection_type = JSONAttrList



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import itertools
import json
import os
import sys
import time
from concurrent.futures import ThreadPoolExecutor

import pytest
from attr_dict_test import AttrDictTest, AttrListTest
from test_json_collection import JSONCollectionTest, TestJSONDict, TestJSONList

from synced_collections.backends.collection_json import (
    BufferedJSONAttrDict,
    BufferedJSONAttrList,
    BufferedJSONDict,
    BufferedJSONList,
    MemoryBufferedJSONAttrDict,
    MemoryBufferedJSONAttrList,
    MemoryBufferedJSONDict,
    MemoryBufferedJSONList,
)
from synced_collections.errors import BufferedError, MetadataError

ON_WINDOWS = sys.platform.startswith("win32") or sys.platform.startswith("cygwin")


class BufferedJSONCollectionTest(JSONCollectionTest):
    def load(self, collection):
        """Load the data corresponding to a SyncedCollection from disk."""
        with open(collection.filename) as f:
            return json.load(f)


class TestBufferedJSONDict(BufferedJSONCollectionTest, TestJSONDict):
    """Tests of buffering JSONDicts."""

    _collection_type = BufferedJSONDict  # type: ignore

    @pytest.fixture
    def synced_collection2(self, tmpdir):
        yield self._collection_type(
            filename=os.path.join(tmpdir, "test2.json"), write_concern=False
        )

    def test_buffered(self, synced_collection, testdata):
        """Test basic per-instance buffering behavior."""
        assert len(synced_collection) == 0
        synced_collection["buffered"] = testdata
        assert "buffered" in synced_collection
        assert synced_collection["buffered"] == testdata
        with synced_collection.buffered:
            assert "buffered" in synced_collection
            assert synced_collection["buffered"] == testdata
            synced_collection["buffered2"] = 1
            assert "buffered2" in synced_collection
            assert synced_collection["buffered2"] == 1
        assert len(synced_collection) == 2
        assert "buffered2" in synced_collection
        assert synced_collection["buffered2"] == 1
        with synced_collection.buffered:
            del synced_collection["buffered"]
            assert len(synced_collection) == 1
            assert "buffered" not in synced_collection
        assert len(synced_collection) == 1
        assert "buffered" not in synced_collection
        assert "buffered2" in synced_collection
        assert synced_collection["buffered2"] == 1

        # Explicitly check that the file has not been changed when buffering.
        raw_dict = synced_collection()
        with synced_collection.buffered:
            synced_collection["buffered3"] = 1
            on_disk_dict = self.load(synced_collection)
            assert "buffered3" not in on_disk_dict
            assert on_disk_dict == raw_dict

        on_disk_dict = self.load(synced_collection)
        assert "buffered3" in on_disk_dict
        assert on_disk_dict == synced_collection

    def test_two_buffered(self, synced_collection, testdata):
        """Test that a non-buffered copy is not modified."""
        synced_collection["buffered"] = testdata
        synced_collection2 = self._collection_type(filename=synced_collection._filename)

        # Check that the non-buffered object is not modified.
        with synced_collection.buffered:
            synced_collection["buffered2"] = 1
            assert "buffered2" not in synced_collection2

    def test_two_buffered_modify_unbuffered(self, synced_collection, testdata):
        """Test that in-memory changes raise errors in buffered mode."""
        synced_collection["buffered"] = testdata
        synced_collection2 = self._collection_type(filename=synced_collection._filename)

        # Check that the non-buffered object is not modified.
        with pytest.raises(MetadataError):
            with synced_collection.buffered:
                synced_collection["buffered2"] = 1
                synced_collection2["buffered2"] = 2
                assert synced_collection["buffered2"] == 1
                synced_collection["buffered2"] = 3
                assert synced_collection2["buffered2"] == 2
                synced_collection2["buffered2"] = 3
                assert synced_collection["buffered2"] == 3

    def test_two_buffered_modify_unbuffered_first(self, synced_collection, testdata):
        synced_collection["buffered"] = testdata
        synced_collection2 = self._collection_type(filename=synced_collection._filename)

        # Check that the non-buffered object is not modified.
        with synced_collection.buffered:
            synced_collection2["buffered2"] = 1
            assert "buffered2" in synced_collection
            synced_collection["buffered2"] = 3
        assert synced_collection == {"buffered": testdata, "buffered2": 3}

    def test_global_buffered(self, synced_collection, testdata):
        assert len(synced_collection) == 0
        synced_collection["buffered"] = testdata
        assert "buffered" in synced_collection
        assert synced_collection["buffered"] == testdata
        with self._collection_type.buffer_backend():
            assert "buffered" in synced_collection
            assert synced_collection["buffered"] == testdata
            synced_collection["buffered2"] = 1
            assert "buffered2" in synced_collection
            assert synced_collection["buffered2"] == 1
        assert len(synced_collection) == 2
        assert "buffered2" in synced_collection
        assert synced_collection["buffered2"] == 1
        with self._collection_type.buffer_backend():
            del synced_collection["buffered"]
            assert len(synced_collection) == 1
            assert "buffered" not in synced_collection
        assert len(synced_collection) == 1
        assert "buffered" not in synced_collection
        assert "buffered2" in synced_collection
        assert synced_collection["buffered2"] == 1

        with pytest.raises(BufferedError):
            with self._collection_type.buffer_backend():
                synced_collection["buffered2"] = 2
                self.store(synced_collection, {"test": 1})
                assert synced_collection["buffered2"] == 2
        assert "test" in synced_collection
        assert synced_collection["test"] == 1

    def test_nested_same_collection(self, synced_collection):
        """Test nesting global buffering."""
        assert len(synced_collection) == 0

        for outer_buffer, inner_buffer in itertools.product(
            [synced_collection.buffered, self._collection_type.buffer_backend()],
            repeat=2,
        ):
            err_msg = (
                f"outer_buffer: {type(outer_buffer).__qualname__}, "
                f"inner_buffer: {type(inner_buffer).__qualname__}"
            )
            synced_collection.reset({"outside": 1})
            with outer_buffer():
                synced_collection["inside_first"] = 2
                with inner_buffer():
                    synced_collection["inside_second"] = 3

                on_disk_dict = self.load(synced_collection)
                assert "inside_first" not in on_disk_dict, err_msg
                assert "inside_second" not in on_disk_dict, err_msg
                assert "inside_first" in synced_collection, err_msg
                assert "inside_second" in synced_collection, err_msg

        assert self.load(synced_collection) == synced_collection

    def test_nested_different_collections(self, synced_collection, synced_collection2):
        """Test nested buffering for different collections."""
        assert len(synced_collection) == 0
        assert len(synced_collection2) == 0

        synced_collection["outside"] = 1
        synced_collection2["outside"] = 1
        with synced_collection.buffered:
            synced_collection["inside_first"] = 2
            on_disk_dict = self.load(synced_collection)
            assert "inside_first" in synced_collection
            assert "inside_first" not in on_disk_dict

            synced_collection2["inside_first"] = 2
            on_disk_dict2 = self.load(synced_collection2)
            assert "inside_first" in synced_collection2
            assert "inside_first" in on_disk_dict2

            with self._collection_type.buffer_backend():
                synced_collection["inside_second"] = 3
                synced_collection2["inside_second"] = 3

                on_disk_dict = self.load(synced_collection)
                assert "inside_second" in synced_collection
                assert "inside_second" not in on_disk_dict
                on_disk_dict2 = self.load(synced_collection2)
                assert "inside_second" in synced_collection2
                assert "inside_second" not in on_disk_dict2

            on_disk_dict = self.load(synced_collection)
            on_disk_dict2 = self.load(synced_collection2)

            assert "inside_first" in synced_collection
            assert "inside_first" not in on_disk_dict

            assert "inside_second" in synced_collection
            assert "inside_second" not in on_disk_dict
            assert "inside_second" in synced_collection2
            assert "inside_second" in on_disk_dict2

        on_disk_dict = self.load(synced_collection)
        on_disk_dict2 = self.load(synced_collection2)

        assert "inside_first" in synced_collection
        assert "inside_first" in on_disk_dict

        assert "inside_second" in synced_collection
        assert "inside_second" in on_disk_dict
        assert "inside_second" in synced_collection2
        assert "inside_second" in on_disk_dict2

    def test_nested_copied_collection(self, synced_collection):
        """Test modifying two collections pointing to the same data."""
        synced_collection2 = self._collection_type(filename=synced_collection._filename)

        assert len(synced_collection) == 0
        assert len(synced_collection2) == 0

        synced_collection["outside"] = 1
        with synced_collection.buffered:
            synced_collection["inside_first"] = 2

            on_disk_dict = self.load(synced_collection)
            assert synced_collection["inside_first"] == 2
            assert "inside_first" not in on_disk_dict

            with self._collection_type.buffer_backend():
                synced_collection["inside_second"] = 3
                synced_collection2["inside_second"] = 4

                on_disk_dict = self.load(synced_collection)
                assert synced_collection["inside_second"] == 4
                assert synced_collection2["inside_second"] == 4
                assert "inside_second" not in on_disk_dict

            on_disk_dict = self.load(synced_collection)
            assert on_disk_dict["inside_second"] == 4

    @pytest.mark.skip("This is an example of unsupported (and undefined) behavior).")
    def test_nested_copied_collection_invalid(self, synced_collection):
        """Test the behavior of invalid modifications of copied objects."""
        synced_collection2 = self._collection_type(filename=synced_collection._filename)

        assert len(synced_collection) == 0
        assert len(synced_collection2) == 0

        synced_collection["outside"] = 1
        finished = False
        with pytest.raises(MetadataError):
            with synced_collection.buffered:
                synced_collection["inside_first"] = 2
                # Modifying synced_collection2 here causes problems. It is
                # unbuffered, so it directly writes to file.  Then, when
                # entering global buffering in the context below,
                # synced_collection2 sees that synced_collection has already
                # saved data for this file to the buffer, so it loads that
                # data, which also means that synced_collection2 becomes
                # associated with the metadata stored when synced_collection
                # entered buffered mode. As a result, when the global buffering
                # exits, we see metadata errors because synced_collection2 lost
                # track of the fact that it saved changes to the file made
                # prior to entering the global buffer. While this case could be
                # given some specific behavior, there's no obvious canonical
                # source of truth here, so we simply choose to skip it
                # altogether.
                synced_collection2["inside_first"] = 3

                on_disk_dict = self.load(synced_collection)
                assert synced_collection["inside_first"] == 2
                assert on_disk_dict["inside_first"] == 3

                with self._collection_type.buffer_backend():
                    synced_collection["inside_second"] = 3
                    synced_collection2["inside_second"] = 4

                    on_disk_dict = self.load(synced_collection)
                    assert synced_collection["inside_second"] == 4
                    assert synced_collection2["inside_second"] == 4
                    assert "inside_second" not in on_disk_dict

                on_disk_dict = self.load(synced_collection)
                assert on_disk_dict["inside_second"] == 4
                # Check that all the checks ran before the assertion failure.
                finished = True
        assert finished

    def test_buffer_flush(self, synced_collection):
        """Test that the buffer gets flushed when enough data is written."""
        original_buffer_capacity = self._collection_type.get_buffer_capacity()
        assert self._collection_type.get_current_buffer_size() == 0

        try:
            self._collection_type.set_buffer_capacity(20)

            # Ensure that the file exists on disk by executing a clear operation so
            # that load operations work as expected.
            assert len(synced_collection) == 0
            synced_collection.clear()

            with synced_collection.buffered:
                synced_collection["foo"] = 1
                assert self._collection_type.get_current_buffer_size() == len(
                    repr(synced_collection)
                )
                assert synced_collection != self.load(synced_collection)

                # Add a long enough value to force a flush.
                synced_collection["bar"] = 100
                assert self._collection_type.get_current_buffer_size() == 0

                # Make sure the file on disk now matches.
                assert synced_collection == self.load(synced_collection)
        finally:
            # Reset buffer capacity for other tests.
            self._collection_type.set_buffer_capacity(original_buffer_capacity)

            # To avoid confusing test failures later, make sure the buffer is
            # truly flushed correctly.
            assert self._collection_type.get_current_buffer_size() == 0

    def multithreaded_buffering_test(self, op, tmpdir):
        """Test that buffering in a multithreaded context is safe for different operations.

        This method encodes the logic for the test, but can be used to test different
        operations on the dict.
        """
        original_buffer_capacity = self._collection_type.get_buffer_capacity()
        try:
            # Choose some arbitrarily low value that will ensure intermittent
            # forced buffer flushes.
            new_buffer_capacity = 20
            self._collection_type.set_buffer_capacity(new_buffer_capacity)

            with self._collection_type.buffer_backend():
                num_dicts = 100
                dicts = []
                dict_data = []
                for i in range(num_dicts):
                    fn = os.path.join(tmpdir, f"test_dict{i}.json")
                    dicts.append(self._collection_type(filename=fn))
                    dict_data.append({str(j): j for j in range(i)})

                num_threads = 10
                try:
                    with ThreadPoolExecutor(max_workers=num_threads) as executor:
                        list(executor.map(op, dicts, dict_data))
                except KeyError as e:
                    raise RuntimeError(
                        "Buffering in parallel failed due to different threads "
                        "simultaneously modifying the buffer."
                    ) from e

                # First validate inside buffer.
                assert all(dicts[i] == dict_data[i] for i in range(num_dicts))
            # Now validate outside buffer.
            assert all(dicts[i] == dict_data[i] for i in range(num_dicts))
        finally:
            # Reset buffer capacity for other tests in case this fails.
            self._collection_type.set_buffer_capacity(original_buffer_capacity)

            # To avoid confusing test failures later, make sure the buffer is
            # truly flushed correctly.
            assert self._collection_type.get_current_buffer_size() == 0

    @pytest.mark.skipif(
        ON_WINDOWS,
        reason=(
            "The JSONCollection cannot be safely used in multithreaded settings "
            "on Windows due to https://bugs.python.org/issue46003."
        ),
    )
    def test_multithreaded_buffering_setitem(self, tmpdir):
        """Test setitem in a multithreaded buffering context."""

        def setitem_dict(sd, data):
            for k, v in data.items():
                sd[k] = v

        self.multithreaded_buffering_test(setitem_dict, tmpdir)

    @pytest.mark.skipif(
        ON_WINDOWS,
        reason=(
            "The JSONCollection cannot be safely used in multithreaded settings "
            "on Windows due to https://bugs.python.org/issue46003."
        ),
    )
    def test_multithreaded_buffering_update(self, tmpdir):
        """Test update in a multithreaded buffering context."""

        def update_dict(sd, data):
            sd.update(data)

        self.multithreaded_buffering_test(update_dict, tmpdir)

    @pytest.mark.skipif(
        ON_WINDOWS,
        reason=(
            "The JSONCollection cannot be safely used in multithreaded settings "
            "on Windows due to https://bugs.python.org/issue46003."
        ),
    )
    def test_multithreaded_buffering_reset(self, tmpdir):
        """Test reset in a multithreaded buffering context."""

        def reset_dict(sd, data):
            sd.reset(data)

        self.multithreaded_buffering_test(reset_dict, tmpdir)

    @pytest.mark.skipif(
        ON_WINDOWS,
        reason=(
            "The JSONCollection cannot be safely used in multithreaded settings "
            "on Windows due to https://bugs.python.org/issue46003."
        ),
    )
    def test_multithreaded_buffering_clear(self, tmpdir):
        """Test clear in a multithreaded buffering context.

        Since clear requires ending up with an empty dict, it's easier to
        write a separate test from the others.
        """
        original_buffer_capacity = self._collection_type.get_buffer_capacity()
        try:
            # Choose some arbitrarily low value that will ensure intermittent
            # forced buffer flushes.
            new_buffer_capacity = 20
            self._collection_type.set_buffer_capacity(new_buffer_capacity)

            # Initialize the data outside the buffered context so that it's
            # already present on disk for testing both.
            num_dicts = 100
            dicts = []
            for i in range(num_dicts):
                fn = os.path.join(tmpdir, f"test_dict{i}.json")
                dicts.append(self._collection_type(filename=fn))
                dicts[-1].update({str(j): j for j in range(i)})

            with self._collection_type.buffer_backend():
                num_threads = 10
                try:
                    with ThreadPoolExecutor(max_workers=num_threads) as executor:
                        list(executor.map(lambda sd: sd.clear(), dicts))
                except KeyError as e:
                    raise RuntimeError(
                        "Buffering in parallel failed due to different threads "
                        "simultaneously modifying the buffer."
                    ) from e

                # First validate inside buffer.
                assert all(not dicts[i] for i in range(num_dicts))
            # Now validate outside buffer.
            assert all(not dicts[i] for i in range(num_dicts))
        finally:
            # Reset buffer capacity for other tests in case this fails.
            self._collection_type.set_buffer_capacity(original_buffer_capacity)

            # To avoid confusing test failures later, make sure the buffer is
            # truly flushed correctly.
            assert self._collection_type.get_current_buffer_size() == 0

    @pytest.mark.skipif(
        ON_WINDOWS,
        reason=(
            "The JSONCollection cannot be safely used in multithreaded settings "
            "on Windows due to https://bugs.python.org/issue46003."
        ),
    )
    def test_multithreaded_buffering_load(self, tmpdir):
        """Test loading data in a multithreaded buffering context.

        This test is primarily for verifying that multithreaded buffering does
        not lead to concurrency errors in flushing data from the buffer due to
        too many loads. This test is primarily for buffering methods with a maximum
        capacity, even for read-only operations.
        """
        original_buffer_capacity = self._collection_type.get_buffer_capacity()
        try:
            # Choose some arbitrarily low value that will ensure intermittent
            # forced buffer flushes.
            new_buffer_capacity = 1000
            self._collection_type.set_buffer_capacity(new_buffer_capacity)

            # Must initialize the data outside the buffered context so that
            # we only execute read operations inside the buffered context.
            num_dicts = 100
            dicts = []
            for i in range(num_dicts):
                fn = os.path.join(tmpdir, f"test_dict{i}.json")
                dicts.append(self._collection_type(filename=fn))
                # Go to i+1 so that every dict contains the 0 element.
                dicts[-1].update({str(j): j for j in range(i + 1)})

            with self._collection_type.buffer_backend():
                num_threads = 100
                try:
                    with ThreadPoolExecutor(max_workers=num_threads) as executor:
                        list(executor.map(lambda sd: sd["0"], dicts * 5))
                except KeyError as e:
                    raise RuntimeError(
                        "Buffering in parallel failed due to different threads "
                        "simultaneously modifying the buffer."
                    ) from e

        finally:
            # Reset buffer capacity for other tests in case this fails.
            self._collection_type.set_buffer_capacity(original_buffer_capacity)

            # To avoid confusing test failures later, make sure the buffer is
            # truly flushed correctly.
            assert self._collection_type.get_current_buffer_size() == 0

    def test_buffer_first_load(self, synced_collection):
        """Ensure that existing data is preserved if the first load is in buffered mode."""
        fn = synced_collection.filename
        write_concern = self._write_concern

        sc = self._collection_type(fn, write_concern)
        sc["foo"] = 1
        sc["bar"] = 2
        del sc

        sc = self._collection_type(fn, write_concern)
        with sc.buffered:
            sc["foo"] = 3

        assert "bar" in sc

    def test_data_type_with_buffered(self, synced_collection):
        """Make sure that the _data attribute has the right types after buffering."""
        with self._collection_type.buffer_backend():
            synced_collection["foo"] = {"bar": "baz"}

        # This will fail if synced_collection['foo'] is a raw dict.
        synced_collection["foo"]()


class TestBufferedJSONList(BufferedJSONCollectionTest, TestJSONList):
    """Tests of buffering JSONLists."""

    _collection_type = BufferedJSONList  # type: ignore

    def test_buffered(self, synced_collection):
        synced_collection.extend([1, 2, 3])
        assert len(synced_collection) == 3
        assert synced_collection == [1, 2, 3]
        with synced_collection.buffered:
            assert len(synced_collection) == 3
            assert synced_collection == [1, 2, 3]
            synced_collection[0] = 4
            assert len(synced_collection) == 3
            assert synced_collection == [4, 2, 3]
        assert len(synced_collection) == 3
        assert synced_collection == [4, 2, 3]
        with synced_collection.buffered:
            assert len(synced_collection) == 3
            assert synced_collection == [4, 2, 3]
            del synced_collection[0]
            assert len(synced_collection) == 2
            assert synced_collection == [2, 3]
        assert len(synced_collection) == 2
        assert synced_collection == [2, 3]

        # Explicitly check that the file has not been changed when buffering.
        raw_list = synced_collection()
        with synced_collection.buffered:
            synced_collection.append(10)
            on_disk_list = self.load(synced_collection)
            assert 10 not in on_disk_list
            assert on_disk_list == raw_list

        on_disk_list = self.load(synced_collection)
        assert 10 in on_disk_list
        assert on_disk_list == synced_collection

    def test_global_buffered(self, synced_collection):
        assert len(synced_collection) == 0
        with self._collection_type.buffer_backend():
            synced_collection.reset([1, 2, 3])
            assert len(synced_collection) == 3
        assert len(synced_collection) == 3
        assert synced_collection == [1, 2, 3]
        with self._collection_type.buffer_backend():
            assert len(synced_collection) == 3
            assert synced_collection == [1, 2, 3]
            synced_collection[0] = 4
            assert len(synced_collection) == 3
            assert synced_collection == [4, 2, 3]
        assert len(synced_collection) == 3
        assert synced_collection == [4, 2, 3]

        # metacheck failure
        with pytest.raises(BufferedError):
            with self._collection_type.buffer_backend():
                synced_collection.reset([1])
                assert synced_collection == [1]
                # Unfortunately the resolution of os.stat is
                # platform dependent and may not always be
                # high enough for our check to work. Since
                # this unit test is artificially simple we
                # must add some amount of minimum waiting time
                # to ensure that the change in time will be
                # detected.
                time.sleep(0.01)
                self.store(synced_collection, [1, 2, 3])
                assert synced_collection == [1]
        assert len(synced_collection) == 3
        assert synced_collection == [1, 2, 3]

    def multithreaded_buffering_test(self, op, requires_init, tmpdir):
        """Test that buffering in a multithreaded context is safe for different operations.

        This method encodes the logic for the test, but can be used to test different
        operations on the list.
        """
        original_buffer_capacity = self._collection_type.get_buffer_capacity()
        try:
            # Choose some arbitrarily low value that will ensure intermittent
            # forced buffer flushes.
            new_buffer_capacity = 20
            self._collection_type.set_buffer_capacity(new_buffer_capacity)

            num_lists = 100
            lists = []
            list_data = []
            for i in range(num_lists):
                # Initialize data with zeros, but prepare other data for
                # updating in place.
                fn = os.path.join(tmpdir, f"test_list{i}.json")
                lists.append(self._collection_type(filename=fn))
                list_data.append([j for j in range(i)])
                if requires_init:
                    lists[-1].extend([0 for j in range(i)])

            with self._collection_type.buffer_backend():
                num_threads = 10
                try:
                    with ThreadPoolExecutor(max_workers=num_threads) as executor:
                        list(executor.map(op, lists, list_data))
                except KeyError as e:
                    raise RuntimeError(
                        "Buffering in parallel failed due to different threads "
                        "simultaneously modifying the buffer."
                    ) from e

                # First validate inside buffer.
                assert all(lists[i] == list_data[i] for i in range(num_lists))
            # Now validate outside buffer.
            assert all(lists[i] == list_data[i] for i in range(num_lists))
        finally:
            # Reset buffer capacity for other tests in case this fails.
            self._collection_type.set_buffer_capacity(original_buffer_capacity)

            # To avoid confusing test failures later, make sure the buffer is
            # truly flushed correctly.
            assert self._collection_type.get_current_buffer_size() == 0

    @pytest.mark.skipif(
        ON_WINDOWS,
        reason=(
            "The JSONCollection cannot be safely used in multithreaded settings "
            "on Windows due to https://bugs.python.org/issue46003."
        ),
    )
    def test_multithreaded_buffering_setitem(self, tmpdir):
        """Test setitem in a multithreaded buffering context."""

        def setitem_list(sd, data):
            for i, val in enumerate(data):
                sd[i] = val

        self.multithreaded_buffering_test(setitem_list, True, tmpdir)

    @pytest.mark.skipif(
        ON_WINDOWS,
        reason=(
            "The JSONCollection cannot be safely used in multithreaded settings "
            "on Windows due to https://bugs.python.org/issue46003."
        ),
    )
    def test_multithreaded_buffering_extend(self, tmpdir):
        """Test extend in a multithreaded buffering context."""

        def extend_list(sd, data):
            sd.extend(data)

        self.multithreaded_buffering_test(extend_list, False, tmpdir)

    @pytest.mark.skipif(
        ON_WINDOWS,
        reason=(
            "The JSONCollection cannot be safely used in multithreaded settings "
            "on Windows due to https://bugs.python.org/issue46003."
        ),
    )
    def test_multithreaded_buffering_append(self, tmpdir):
        """Test append in a multithreaded buffering context."""

        def append_list(sd, data):
            for val in data:
                sd.append(val)

        self.multithreaded_buffering_test(append_list, False, tmpdir)

    @pytest.mark.skipif(
        ON_WINDOWS,
        reason=(
            "The JSONCollection cannot be safely used in multithreaded settings "
            "on Windows due to https://bugs.python.org/issue46003."
        ),
    )
    def test_multithreaded_buffering_load(self, tmpdir):
        """Test loading data in a multithreaded buffering context.

        This test is primarily for verifying that multithreaded buffering does
        not lead to concurrency errors in flushing data from the buffer due to
        too many loads. This test is primarily for buffering methods with a maximum
        capacity, even for read-only operations.
        """
        original_buffer_capacity = self._collection_type.get_buffer_capacity()
        try:
            # Choose some arbitrarily low value that will ensure intermittent
            # forced buffer flushes.
            new_buffer_capacity = 1000
            self._collection_type.set_buffer_capacity(new_buffer_capacity)

            # Must initialize the data outside the buffered context so that
            # we only execute read operations inside the buffered context.
            num_lists = 100
            lists = []
            for i in range(num_lists):
                fn = os.path.join(tmpdir, f"test_list{i}.json")
                lists.append(self._collection_type(filename=fn))
                # Go to i+1 so that every list contains the 0 element.
                lists[-1].extend([j for j in range(i + 1)])

            with self._collection_type.buffer_backend():
                num_threads = 100
                try:
                    with ThreadPoolExecutor(max_workers=num_threads) as executor:
                        list(executor.map(lambda sd: sd[0], lists * 5))
                except KeyError as e:
                    raise RuntimeError(
                        "Buffering in parallel failed due to different threads "
                        "simultaneously modifying the buffer."
                    ) from e
        finally:
            # Reset buffer capacity for other tests in case this fails.
            self._collection_type.set_buffer_capacity(original_buffer_capacity)

            # To avoid confusing test failures later, make sure the buffer is
            # truly flushed correctly.
            assert self._collection_type.get_current_buffer_size() == 0


class TestMemoryBufferedJSONDict(TestBufferedJSONDict):
    """Tests of MemoryBufferedJSONDicts."""

    _collection_type = MemoryBufferedJSONDict  # type: ignore

    def test_buffer_flush(self, synced_collection, synced_collection2):
        """Test that the buffer gets flushed when enough data is written."""
        original_buffer_capacity = self._collection_type.get_buffer_capacity()

        assert self._collection_type.get_current_buffer_size() == 0
        self._collection_type.set_buffer_capacity(1)

        # Ensure that the file exists on disk by executing a clear operation so
        # that load operations work as expected.
        assert len(synced_collection) == 0
        assert len(synced_collection2) == 0
        synced_collection.clear()
        synced_collection2.clear()

        with self._collection_type.buffer_backend():
            synced_collection["foo"] = 1
            assert self._collection_type.get_current_buffer_size() == 1
            assert synced_collection != self.load(synced_collection)

            # This buffering mode is based on the number of files buffered, so
            # we need to write to the second collection.
            assert "bar" not in synced_collection2

            # Simply loading the second collection into memory shouldn't
            # trigger a flush, because it hasn't been modified and we flush
            # based on the total number of modifications.
            assert synced_collection != self.load(synced_collection)

            # Modifying the second collection should exceed buffer capacity and
            # trigger a flush.
            synced_collection2["bar"] = 2
            assert synced_collection == self.load(synced_collection)
            assert synced_collection2 == self.load(synced_collection2)

        # Reset buffer capacity for other tests.
        self._collection_type.set_buffer_capacity(original_buffer_capacity)


class TestMemoryBufferedJSONList(TestBufferedJSONList):
    """Tests of MemoryBufferedJSONLists."""

    _collection_type = MemoryBufferedJSONList  # type: ignore


class TestBufferedJSONDictWriteConcern(TestBufferedJSONDict):
    _write_concern = True


class TestBufferedJSONListWriteConcern(TestBufferedJSONList):
    _write_concern = True


class TestBufferedJSONAttrDict(TestBufferedJSONDict, AttrDictTest):
    _collection_type = BufferedJSONAttrDict  # type: ignore


class TestBufferedJSONAttrList(TestBufferedJSONList, AttrListTest):
    _collection_type = BufferedJSONAttrList  # type: ignore


class TestMemoryBufferedJSONAttrDict(TestMemoryBufferedJSONDict, AttrDictTest):
    _collection_type = MemoryBufferedJSONAttrDict  # type: ignore


class TestMemoryBufferedJSONAttrList(TestMemoryBufferedJSONList, AttrListTest):
    _collection_type = MemoryBufferedJSONAttrList  # type: ignore



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Defines the buffering protocol for synced collections.

In addition to defining the buffering protocol for synced collections in
:class:`~.BufferedCollection`, this subpackage also defines a number of
supported buffering implementations. No buffers are imported by default. Users
should import desired buffers as needed.
"""
from typing import List as _List

__all__: _List[str] = []



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import json
import uuid

import pytest
from synced_collection_test import SyncedDictTest, SyncedListTest

from synced_collections.backends.collection_redis import RedisDict, RedisList

try:
    import redis

    try:
        # Try to connect to server
        redis_client = redis.Redis()
        test_key = str(uuid.uuid4())
        redis_client.set(test_key, 0)
        if redis_client.get(test_key) != b"0":  # Redis stores data as bytes
            raise RuntimeError("Cache access check failed.")
        redis_client.delete(test_key)
        REDIS = True
    except (redis.exceptions.ConnectionError, RuntimeError):
        REDIS = False
except ImportError:
    REDIS = False


class RedisCollectionTest:
    _key = "test"

    def store(self, synced_collection, data):
        synced_collection.client.set(synced_collection.key, json.dumps(data).encode())

    @pytest.fixture
    def synced_collection(self, request):
        request.addfinalizer(redis_client.flushall)
        yield self._collection_type(key=self._key, client=redis_client)

    @pytest.fixture
    def synced_collection_positional(self, request):
        """Fixture that initializes the object using positional arguments."""
        request.addfinalizer(redis_client.flushall)
        yield self._collection_type(redis_client, self._key)

    def test_key(self, synced_collection):
        assert synced_collection.key == self._key


@pytest.mark.skipif(
    not REDIS, reason="test requires the redis package and running redis-server"
)
class TestRedisDict(RedisCollectionTest, SyncedDictTest):
    _collection_type = RedisDict


@pytest.mark.skipif(
    not REDIS, reason="test requires the redis package and running redis-server"
)
class TestRedisList(RedisCollectionTest, SyncedListTest):
    _collection_type = RedisList



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Buffering for file-based backends using a serialized buffer.

The buffering method implemented here involves a single buffer of serialized
data. All collections in buffered mode encode their data into this buffer on save
and decode from it on load.
"""

import hashlib
import json

from ..errors import MetadataError
from ..utils import SyncedCollectionJSONEncoder
from .file_buffered_collection import FileBufferedCollection


class SerializedFileBufferedCollection(FileBufferedCollection):
    """A :class:`~.FileBufferedCollection` based on a serialized data store.

    This class extends the :class:`~.FileBufferedCollection` and implements a
    concrete storage mechanism in which data is encoded (by default, into JSON)
    and stored into a buffer. This buffer functions as a central data store for
    all collections and is a synchronization point for various collections
    pointing to the same underlying file. This serialization method may be a
    bottleneck in some applications; see the Warnings section for more information.

    The buffer size and capacity for this class is measured in the total number
    of bytes stored in the buffer that correspond to file data. This is *not*
    the total size of the buffer, which also contains additional information
    like the hash of the data and the file metadata (which are used for
    integrity checks), but it is the relevant metric for users.

    .. note::
        Important note for subclasses: This class should be inherited before
        any other collections. This requirement is due to the extensive use of
        multiple inheritance: since this class is designed to be combined with
        other :class:`~.SyncedCollection` types without making those types aware
        of buffering behavior, it transparently hooks into the initialization
        process, but this is dependent on its constructor being called before
        those of other classes.

    **Thread safety**

    This buffering method is thread safe. This thread safety is independent of the
    safety of an individual collection backend; the backend must support thread
    safe writes to the underlying resource in order for a buffered version using
    this class to be thread safe for general use. The thread safety guaranteed
    by this class only concerns buffer reads, writes, and flushes. All these
    operations are serialized because there is no way to prevent one collection
    from triggering a flush while another still thinks its data is in the cache.

    Parameters
    ----------
    filename : str, optional
        The filename of the associated JSON file on disk (Default value = None).

    Warnings
    --------
    - Although it can be done safely, in general modifying two different collections
      pointing to the same underlying resource while both are in different buffering
      modes is unsupported and can lead to undefined behavior. This class makes a
      best effort at performing safe modifications, but it is possible to construct
      nested buffered contexts for different objects that can lead to an invalid
      buffer state, or even situations where there is no obvious indicator of what
      is the canonical source of truth. In general, if you need multiple objects
      pointing to the same resource, it is **strongly** recommeneded to work with
      both of them in identical buffering states at all times.
    - The overhead of this buffering method is quite high due to the constant
      encoding and decoding of data. For performance-critical applications where
      memory is not highly constrained and virtual memory limits are absent, the
      :class:`~.SharedMemoryFileBufferedCollection` may be more appropriate.
    - Due to the possibility of read operations triggering a flush, the
      contents of the buffer may be invalidated on loads as well. To prevent this
      even nominally read-only operations are serialized. As a result, although
      this class is thread safe, it will effectively serialize all operations and
      will therefore not be performant.

    """

    _BUFFER_CAPACITY = 32 * 2**20  # 32 MB

    def _flush(self, force=False):
        """Save buffered changes to the underlying file.

        Parameters
        ----------
        force : bool
            If True, force a flush even in buffered mode (defaults to False). This
            parameter is used when the buffer is filled to capacity.

        Raises
        ------
        MetadataError
            If any file is detected to have changed on disk since it was
            originally loaded into the buffer and modified.

        """
        # Different files in the buffer can be safely flushed simultaneously,
        # but a given file can only be flushed on one thread at once.
        with self._buffer_lock:
            if not self._is_buffered or force:
                try:
                    cached_data = type(self)._buffer[self._filename]
                except KeyError:
                    # There are valid reasons for nothing to be in the cache (the
                    # object was never actually accessed during global buffering,
                    # multiple collections pointing to the same file, etc).
                    return
                else:
                    blob = self._encode(self._data)

                    # If the contents have not been changed since the initial read,
                    # we don't need to rewrite it.
                    try:
                        if self._hash(blob) != cached_data["hash"]:
                            # Validate that the file hasn't been changed by
                            # something else.
                            if cached_data["metadata"] != self._get_file_metadata():
                                raise MetadataError(
                                    self._filename, cached_data["contents"]
                                )
                            self._update(self._decode(cached_data["contents"]))
                            self._save_to_resource()
                    finally:
                        # Whether or not an error was raised, the cache must be
                        # cleared to ensure a valid final buffer state.
                        del type(self)._buffer[self._filename]
                        data_size = len(cached_data["contents"])
                        type(self)._CURRENT_BUFFER_SIZE -= data_size

    @staticmethod
    def _hash(blob):
        """Calculate and return the md5 hash value for the file data.

        Parameters
        ----------
        blob : bytes
            Byte literal to be hashed.

        Returns
        -------
        str
            The md5 hash of the input bytes.

        """
        if blob is not None:
            m = hashlib.md5()
            m.update(blob)
            return m.hexdigest()

    @staticmethod
    def _encode(data):
        """Encode the data into a serializable form.

        This method assumes JSON-serializable data, but subclasses can override
        this hook method to change the encoding behavior as needed.

        Parameters
        ----------
        data : collections.abc.Collection
            Any collection type that can be encoded.

        Returns
        -------
        bytes
            The underlying encoded data.

        """
        return json.dumps(data, cls=SyncedCollectionJSONEncoder).encode()

    @staticmethod
    def _decode(blob):
        """Decode serialized data.

        This method assumes JSON-serializable data, but subclasses can override
        this hook method to change the encoding behavior as needed.

        Parameters
        ----------
        blob : bytes
            Byte literal to be decoded.

        Returns
        -------
        data : collections.abc.Collection
            The decoded data in the appropriate base collection type.

        """
        return json.loads(blob.decode())

    def _save_to_buffer(self):
        """Store data in buffer.

        See :meth:`~._initialize_data_in_buffer` for details on the data stored
        in the buffer and the integrity checks performed.
        """
        type(self)._buffered_collections[id(self)] = self

        # Since one object could write to the buffer and trigger a flush while
        # another object was found in the buffer and attempts to proceed
        # normally, we have to serialize this whole block. In theory we might
        # be safe without it because the only operations that should reach this
        # point without already being locked are destructive operations (clear,
        # reset) that don't use the :meth:`_load_and_save` context, and for
        # those the writes will be automatically serialized because Python
        # dicts are thread-safe because of the GIL. However, it's best not to
        # depend on the thread-safety of built-in containers.
        with self._buffer_lock:
            if self._filename in type(self)._buffer:
                # Always track all instances pointing to the same data.
                blob = self._encode(self._data)
                cached_data = type(self)._buffer[self._filename]
                buffer_size_change = len(blob) - len(cached_data["contents"])
                type(self)._CURRENT_BUFFER_SIZE += buffer_size_change
                cached_data["contents"] = blob
            else:
                # The only methods that could safely call sync without a load are
                # destructive operations like `reset` or `clear` that completely
                # wipe out previously existing data. Therefore, the safest choice
                # for ensuring consistency of the buffer is to modify the stored
                # hash (which is used for the consistency check) with the hash of
                # the current data on disk. _initialize_data_in_buffer always uses
                # the current metadata, so the only extra work here is to modify
                # the hash after it's called (since it uses self._to_base() to get
                # the data to initialize the cache with).
                self._initialize_data_in_buffer()
                disk_data = self._load_from_resource()
                type(self)._buffer[self._filename]["hash"] = self._hash(
                    self._encode(disk_data)
                )

            if type(self)._CURRENT_BUFFER_SIZE > type(self)._BUFFER_CAPACITY:
                type(self)._flush_buffer(force=True)

    def _load_from_buffer(self):
        """Read data from buffer.

        See :meth:`~._initialize_data_in_buffer` for details on the data stored
        in the buffer and the integrity checks performed.

        Returns
        -------
        Collection
            A collection of the same base type as the SyncedCollection this
            method is called for, corresponding to data loaded from the
            underlying file.

        """
        with self._buffer_lock:
            super()._load_from_buffer()

            # Load from buffer. This has to happen inside the locked context
            # because otherwise the data could be flushed from the buffer by
            # another thread.
            blob = type(self)._buffer[self._filename]["contents"]

        if type(self)._CURRENT_BUFFER_SIZE > type(self)._BUFFER_CAPACITY:
            type(self)._flush_buffer(force=True)
        return self._decode(blob)

    def _initialize_data_in_buffer(self):
        """Create the initial entry for the data in the cache.

        Stores the following information:
            - The hash of the data as initially stored in the cache. This hash
              is used to quickly determine whether data has changed when flushing.
            - The metadata provided by :meth:`~._get_file_metadata`. Used to
              check if a file has been modified on disk since it was loaded
              into the buffer.

        This method also increments the current buffer size, which in this class
        is the total number of bytes of data in the buffer.
        """
        blob = self._encode(self._data)
        metadata = self._get_file_metadata()

        type(self)._buffer[self._filename] = {
            "contents": blob,
            "hash": self._hash(blob),
            "metadata": metadata,
        }
        type(self)._CURRENT_BUFFER_SIZE += len(
            type(self)._buffer[self._filename]["contents"]
        )

    @classmethod
    def _flush_buffer(cls, force=False):
        """Flush the data in the file buffer.

        Parameters
        ----------
        force : bool
            If True, force a flush even in buffered mode (defaults to False). This
            parameter is used when the buffer is filled to capacity.

        Returns
        -------
        dict
            Mapping of filename and errors occured during flushing data.

        Raises
        ------
        BufferedError
            If there are any issues with flushing the data.

        """
        return super()._flush_buffer(force=force, retain_in_force=False)



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import pytest
from synced_collection_test import SyncedDictTest, SyncedListTest

from synced_collections.backends.collection_zarr import ZarrDict, ZarrList

try:
    import numcodecs  # zarr depends on numcodecs
    import zarr

    ZARR = True
except ImportError:
    ZARR = False


class ZarrCollectionTest:
    _name = "test"

    def store(self, synced_collection, data):
        dataset = synced_collection.group.require_dataset(
            "test",
            overwrite=True,
            shape=1,
            dtype="object",
            object_codec=numcodecs.JSON(),
        )
        dataset[0] = data

    @pytest.fixture
    def synced_collection(self, tmpdir):
        yield self._collection_type(
            name=self._name, group=zarr.group(zarr.DirectoryStore(tmpdir))
        )

    @pytest.fixture
    def synced_collection_positional(self, tmpdir):
        """Fixture that initializes the object using positional arguments."""
        yield self._collection_type(zarr.group(zarr.DirectoryStore(tmpdir)), self._name)

    def test_name(self, synced_collection):
        assert synced_collection.name == self._name


@pytest.mark.skipif(not ZARR, reason="test requires the zarr package")
class TestZarrDict(ZarrCollectionTest, SyncedDictTest):
    _collection_type = ZarrDict


@pytest.mark.skipif(not ZARR, reason="test requires the zarr package")
class TestZarrList(ZarrCollectionTest, SyncedListTest):
    _collection_type = ZarrList



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implements a Redis :class:`~.SyncedCollection` backend."""
import json

from .. import SyncedCollection, SyncedDict, SyncedList
from ..utils import SyncedCollectionJSONEncoder
from ..validators import json_format_validator, require_string_key


class RedisCollection(SyncedCollection):
    r"""A :class:`~.SyncedCollection` that synchronizes with a Redis database.

    This backend stores data in Redis by associating it with the provided key.

    **Thread safety**

    The :class:`RedisCollection` is not thread-safe.

    Parameters
    ----------
    client : redis.Redis
        The Redis client used to persist data.
    key : str
        The key associated with this collection in the Redis database.
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    """

    _backend = __name__  # type: ignore

    # Redis collection relies on JSON-serialization for the data.
    _validators = (json_format_validator,)

    def __init__(self, client=None, key=None, *args, **kwargs):
        super().__init__(**kwargs)
        self._client = client
        self._key = key

    def _load_from_resource(self):
        """Load the data from a Redis database.

        Returns
        -------
        Collection or None
            An equivalent unsynced collection satisfying :meth:`~.is_base_type` that
            contains the data in the Redis database. Will return None if no data
            was found in the Redis database.

        """
        blob = self._client.get(self._key)
        return None if blob is None else json.loads(blob)

    def _save_to_resource(self):
        """Write the data to a Redis database."""
        self._client.set(
            self._key, json.dumps(self, cls=SyncedCollectionJSONEncoder).encode()
        )

    @property
    def client(self):
        """`redis.Redis`: The Redis client used to store the data."""
        return self._client

    @property
    def key(self):
        """str: The key associated with this collection stored in Redis."""
        return self._key

    def __deepcopy__(self, memo):
        # The underlying Redis client cannot be deepcopied.
        raise TypeError("RedisCollection does not support deepcopying.")


class RedisDict(RedisCollection, SyncedDict):
    r"""A dict-like data structure that synchronizes with a persistent Redis database.

    Examples
    --------
    >>> doc = RedisDict('data')
    >>> doc['foo'] = "bar"
    >>> assert doc['foo'] == "bar"
    >>> assert 'foo' in doc
    >>> del doc['foo']
    >>> doc['foo'] = dict(bar=True)
    >>> doc
    {'foo': {'bar': True}}

    Parameters
    ----------
    client : redis.Redis, optional
        A redis client (Default value = None).
    key : str, optional
        The key of the  collection (Default value = None).
    data : :class:`collections.abc.Mapping`, optional
        The initial data passed to :class:`RedisDict`. If ``None``, defaults to
        ``{}`` (Default value = None).
    parent : RedisCollection, optional
        A parent instance of :class:`RedisCollection` or ``None``. If ``None``,
        the collection owns its own data (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    While the :class:`RedisDict` object behaves like a :class:`dict`, there are important
    distinctions to remember. In particular, because operations are reflected
    as changes to an underlying database, copying a :class:`RedisDict` instance may
    exhibit unexpected behavior. If a true copy is required, you should use the
    call operator to get a dictionary representation, and if necessary
    construct a new :class:`RedisDict` instance.

    """

    _validators = (require_string_key,)

    def __init__(self, client=None, key=None, data=None, parent=None, *args, **kwargs):
        super().__init__(
            client=client, key=key, data=data, parent=parent, *args, **kwargs
        )


class RedisList(RedisCollection, SyncedList):
    r"""A list-like data structure that synchronizes with a persistent Redis database.

    Only non-string sequences are supported by this class.

    Examples
    --------
    >>> synced_list = RedisList('data')
    >>> synced_list.append("bar")
    >>> assert synced_list[0] == "bar"
    >>> assert len(synced_list) == 1
    >>> del synced_list[0]


    Parameters
    ----------
    client : redis.Redis, optional
        A Redis client (Default value = None).
    key : str, optional
        The key of the  collection (Default value = None).
    data : non-str :class:`collections.abc.Sequence`, optional
        The initial data passed to :class:`RedisList`. If ``None``, defaults to
        ``[]`` (Default value = None).
    parent : RedisCollection, optional
        A parent instance of :class:`RedisCollection` or ``None``. If ``None``,
        the collection owns its own data (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    While the :class:`RedisList` object behaves like a :class:`list`, there are
    important distinctions to remember. In particular, because operations are
    reflected as changes to an underlying database, copying a
    :class:`RedisList` instance may exhibit unexpected behavior. If a true copy
    is required, you should use the call operator to get a dictionary
    representation, and if necessary construct a new :class:`RedisList`
    instance.
    """

    def __init__(self, client=None, key=None, data=None, parent=None, *args, **kwargs):
        super().__init__(
            client=client, key=key, data=data, parent=parent, *args, **kwargs
        )



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Define various synced data types."""

from .synced_collection import SyncedCollection
from .synced_dict import SyncedDict
from .synced_list import SyncedList

__all__ = ["SyncedCollection", "SyncedDict", "SyncedList"]



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import pytest

from synced_collections.errors import InvalidKeyError, KeyTypeError
from synced_collections.validators import (
    json_format_validator,
    no_dot_in_key,
    require_string_key,
)

try:
    import numpy

    NUMPY = True
except ImportError:
    NUMPY = False


class TestRequireStringKey:
    def test_valid_data(self, testdata):
        test_dict = {}

        key = "valid_str"
        test_dict[key] = testdata
        require_string_key(test_dict)
        assert key in test_dict
        assert test_dict[key] == testdata

    def test_invalid_data(self, testdata):
        # invalid key types
        for key in (0.0, 1.0 + 2.0j, (1, 2, 3), 1, False, None):
            with pytest.raises(KeyTypeError):
                require_string_key({key: testdata})


class TestNoDotInKey:
    def test_valid_data(self, testdata):
        test_dict = {}
        # valid data
        for key in ("valid_str", "another_valid_str"):
            test_dict[key] = testdata
            no_dot_in_key(test_dict)
            assert key in test_dict
            assert test_dict[key] == testdata

    def test_invalid_data(self, testdata):
        # dict key containing dot
        with pytest.raises(InvalidKeyError):
            no_dot_in_key({"a.b": testdata})
        # nested dict key containing dot
        with pytest.raises(InvalidKeyError):
            no_dot_in_key({"nested": {"a.b": 1}})
        # list containing dict
        with pytest.raises(InvalidKeyError):
            no_dot_in_key([{"a.b": 1}])
        # invalid key types
        for key in (0.0, 1.0 + 2.0j, (1, 2, 3)):
            with pytest.raises(KeyTypeError):
                no_dot_in_key({key: testdata})


class TestJSONFormatValidator:
    def test_valid_data(self):
        for data in ("foo", 1, 1.0, True, None, {}, []):
            json_format_validator(data)
            json_format_validator({"test_key": data})
        json_format_validator(("foo", 1, 1.0, True, None, {}, []))

    def test_dict_data(self, testdata):
        for data in ("foo", 1, 1.0, True, None):
            json_format_validator({"test_key": data})
        for key in (0.0, (1, 2, 3)):
            with pytest.raises(KeyTypeError):
                json_format_validator({key: testdata})

    @pytest.mark.skipif(not NUMPY, reason="test requires the numpy package")
    def test_numpy_data(self):
        data = numpy.random.rand(3, 4)
        json_format_validator(data)
        json_format_validator(numpy.float_(3.14))
        # numpy data as dict value
        json_format_validator({"test": data})
        json_format_validator({"test": numpy.float_(1.0)})
        # numpy data in list
        json_format_validator([data, numpy.float_(1.0), 1, "test"])

    def test_invalid_data(self):
        class A:
            pass

        invalid_data = (1.0 + 2.0j, A())
        for data in invalid_data:
            with pytest.raises(TypeError):
                json_format_validator(data)
        # invalid data as dict value
        for data in invalid_data:
            with pytest.raises(TypeError):
                json_format_validator({"test": data})
        # invalid data in tuple
        with pytest.raises(TypeError):
            json_format_validator(invalid_data)

    @pytest.mark.skipif(not NUMPY, reason="test requires the numpy package")
    def test_numpy_invalid_data(self):
        # complex data
        data = numpy.complex128(1 + 2j)
        with pytest.raises(TypeError):
            json_format_validator(data)
        # complex data in ndarray
        data = numpy.asarray([1, 2, 1j, 1 + 2j])
        with pytest.raises(TypeError):
            json_format_validator(data)



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Define common utilities."""

from json import JSONEncoder
from typing import Any, Dict

from .numpy_utils import _convert_numpy, _is_numpy_scalar


class AbstractTypeResolver:
    """Mapping between recognized types and their abstract parents.

    Synced collections are heavily reliant on checking the types of objects to
    determine the appropriate type of behavior in various scenarios. For maximum
    generality, most of these checks use the ABCs defined in :py:mod:`collections.abc`.
    The price of this flexibility is that `isinstance` checks with these classes
    are very slow because the ``__instancecheck__`` hooks are implemented in pure
    Python and require checking many different cases.

    Rather than attempting to directly optimize this behavior, this class provides
    a workaround by which we can amortize the cost of type checks. Given a set
    of types that must be resolved and a way to identify each of these (which
    may be expensive), it maintains a local cache of all instances of a given
    type that have previously been observed. This reduces the cost of type checking
    to a simple ``dict`` lookup, except for the first time a new type is observed.

    Parameters
    ----------
    abstract_type_identifiers : Mapping
        A mapping from a string identifier for a group of types (e.g. ``"MAPPING"``)
        to a callable that can be used to identify that type. Due to insertion order
        guarantees of dictionaries in Python>=3.6 (officially 3.7), it may be beneficial
        to order this dictionary with the most frequently occuring types first.
        However, unless users have many different concrete types implementing
        the same abstract interface (e.g. many Mapping types identified via
        ``isinstance(obj, Mapping)``), any performance gain should be negligible
        since the callables will only be executed once per type.
    cache_blocklist : Sequence, optional
        A sequence of string identifiers from ``abstract_type_identifiers`` that
        should not be cached. If there are cases where objects of the same type
        would be classified into separate groups based on the callables in
        ``abstract_type_identifiers``, this argument allows users to specify that
        this type should not be cached. This argument should be used sparingly
        because performance will quickly degrade if many calls to
        :meth:`get_type` are with types that cannot be cached. The identifiers
        (keys in ``abstract_type_identifiers``) corresponding to elements of the
        blocklist should be placed first in the ``abstract_type_identifiers``
        dictionary since they will never be cached and are therefore the most
        likely callables to be used repeatedly (Default value = None).

    Attributes
    ----------
    abstract_type_identifiers : Dict[str, Callable[Any, bool]]
        A mapping from string identifiers for an abstract type to callables that
        accepts an object and returns True if the object is of the key type and
        False if not.
    type_map : Dict[Type, str]
        A mapping from concrete types to the corresponding named abstract type
        from :attr:`~.abstract_type_identifiers`.

    """

    def __init__(self, abstract_type_identifiers, cache_blocklist=None):
        self.abstract_type_identifiers = abstract_type_identifiers
        self.type_map = {}
        self.cache_blocklist = cache_blocklist if cache_blocklist is not None else ()

    def get_type(self, obj):
        """Get the type string corresponding to this data type.

        Parameters
        ----------
        obj : Any
            Any object whose type to check

        Returns
        -------
        str
            The name of the type, where valid types are the keys of the dict
            argument to the constructor. If the object's type cannot be identified,
            will return ``None``.

        """
        obj_type = type(obj)
        enum_type = None
        try:
            enum_type = self.type_map[obj_type]
        except KeyError:
            for data_type, id_func in self.abstract_type_identifiers.items():
                if id_func(obj):
                    enum_type = data_type
                    break
            if obj_type not in self.cache_blocklist:
                self.type_map[obj_type] = enum_type

        return enum_type


def default(o: Any) -> Dict[str, Any]:  # noqa: D102
    """Get a JSON-serializable version of compatible types.

    This function is suitable for use with JSON-serialization tools as a way to
    serialize :class:`~.SyncedCollection` objects and NumPy arrays.  It will
    attempt to obtain a JSON-serializable representation of an object that is
    otherwise not serializable by attempting to access its ``_data`` attribute.


    Warnings
    --------
    - JSON encoding of numpy arrays is not invertible; once encoded, reloading
      the data will result in converting arrays to lists and numpy numbers into
      ints or floats.
    - This function assumes that the in-memory data for a SyncedCollection is
      up-to-date. If the data has been changed on disk without updating the
      collection, or if this function is used to serialize the data before any
      method is invoked that would load the data from disk, the resulting
      serialized data may be incorrect.

    """
    # NumPy converters return the data unchanged.
    converted_o = _convert_numpy(o)

    # NumPy arrays will be converted to lists, then recursively parsed by the
    # JSON encoder, so we only have to handle the case where we have a scalar
    # type at the bottom level that can't be converted to a Python scalar.
    if _is_numpy_scalar(converted_o):
        raise ValueError(
            "In order for a NumPy type to be JSON-encoded, it must have a corresponding "
            "Python type. All other types, such as NumPy extended-precision types, must "
            "be converted by the user. Note that the existence of a corresponding type "
            "is a necessary but not sufficient condition because not all Python types "
            "can be JSON-encoded. For instance, complex numpy values can be converted "
            "to Python complex values, but these still cannot be JSON-encoded."
        )

    if converted_o is o:
        try:
            return o._data
        except AttributeError as e:
            raise TypeError from e
    else:
        return converted_o


class SyncedCollectionJSONEncoder(JSONEncoder):
    """A :class:`json.JSONEncoder` that handles objects encodeable using :func:`~.default`.

    Warnings
    --------
    - JSON encoding of numpy arrays is not invertible; once encoded, reloading
      the data will result in converting arrays to lists and numpy numbers into
      ints or floats.
    - This class assumes that the in-memory data for a SyncedCollection is
      up-to-date. If the data has been changed on disk without updating the
      collection, or if this class is used to serialize the data before any
      method of the collection is invoked that would load the data from disk,
      the resulting serialized data may be incorrect.

    """

    def default(self, o: Any) -> Dict[str, Any]:  # noqa: D102
        try:
            return default(o)
        except TypeError:
            # Call the super method, which raises a TypeError if it cannot
            # encode the object.
            return super().default(o)


class _NullContext:
    """A nullary context manager.

    There are various cases where we sometimes want to perform a task within a
    particular context, but at other times we wish to ignore that context. The
    most obvious example is a lock for threading: since
    :class:`~.SyncedCollection` allows multithreading support to be enabled or
    disabled, it is important to be able to write code that is agnostic to
    whether or not a mutex must be acquired prior to executing a task. Locks
    support the context manager protocol and are used in that manner throughout
    the code base, so the most transparent way to disable buffering is to
    create a nullary context manager that can be placed as a drop-in
    replacement for the lock so that all other code can handle this in a
    transparent manner.
    """

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __call__(self):
        """Allow usage of the context in a function-like manner."""
        return self


class _CounterContext:
    """A context manager that maintains a total entry count.

    This class simply contains an internal counter that is incremented on every
    entrance and decremented on every exit. It is also truthy and only evaluates
    to True if the count is greater than 0.
    """

    def __init__(self):
        self._count = 0

    def __enter__(self):
        self._count += 1

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._count -= 1

    def __bool__(self):
        return self._count > 0

    def __call__(self):
        """Allow usage of the context in a function-like manner."""
        return self


class _CounterFuncContext(_CounterContext):
    """A counter that performs some operation whenever the counter hits zero.

    This class maintains a counter, and also accepts an arbitrary nullary
    callable to be executed anytime the context exits and the counter hits zero.
    """

    def __init__(self, func):
        super().__init__()
        self._func = func

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        if not self:
            self._func()



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implements the :class:`SyncedDict`.

This implements a dict-like data structure that also conforms to the
:class:`~.SyncedCollection` API and can be combined with any backend type to
give a dict-like API to a synchronized data structure.
"""

from collections.abc import Mapping, MutableMapping

from ..utils import AbstractTypeResolver
from .synced_collection import SyncedCollection, _sc_resolver

# Identifies mappings, which are the base type for this class.
_mapping_resolver = AbstractTypeResolver(
    {
        "MAPPING": lambda obj: isinstance(obj, Mapping),
    }
)


class SyncedDict(SyncedCollection, MutableMapping):
    r"""Implement the dict data structure along with values access through attributes named as keys.

    The SyncedDict inherits from :class:`~.SyncedCollection`
    and :class:`~collections.abc.MutableMapping`. Therefore, it behaves like a
    :class:`dict`.

    Parameters
    ----------
    data : Mapping, optional
        The initial data to populate the dict. If ``None``, defaults to
        ``{}`` (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warning
    -------
    While the :class:`SyncedDict` object behaves like a :class:`dict`,
    there are important distinctions to remember. In particular, because
    operations are reflected as changes to an underlying backend, copying (even
    deep copying) a :class:`SyncedDict` instance may exhibit unexpected
    behavior. If a true copy is required, you should use the :meth:`~.__call__`
    method to get a :class:`dict` representation, and if necessary construct a
    new :class:`SyncedDict`.
    """

    # The _validate parameter is an optimization for internal use only. This
    # argument will be passed by _from_base whenever an unsynced collection is
    # being recursively converted, ensuring that validation only happens once.
    def __init__(self, data=None, _validate=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if data is None:
            self._data = {}
        else:
            if _validate:
                self._validate(data)
            with self._suspend_sync:
                self._data = {
                    key: self._from_base(data=value, parent=self)
                    for key, value in data.items()
                }

    def _to_base(self):
        """Convert the SyncedDict object to a :class:`dict`.

        Returns
        -------
        dict
            An equivalent raw :class:`dict`.

        """
        converted = {}
        for key, value in self._data.items():
            switch_type = _sc_resolver.get_type(value)
            if switch_type == "SYNCEDCOLLECTION":
                converted[key] = value._to_base()
            else:
                converted[key] = value
        return converted

    @classmethod
    def is_base_type(cls, data):
        """Check whether the data is an instance of mapping.

        Parameters
        ----------
        data : any
            Data to be checked.

        Returns
        -------
        bool

        """
        return _mapping_resolver.get_type(data) == "MAPPING"

    def _update(self, data=None, _validate=False):
        """Update the in-memory representation to match the provided data.

        The purpose of this method is to update the SyncedCollection to match
        the data in the underlying resource.  The result of calling this method
        should be that ``self == data``. The reason that this method is
        necessary is that SyncedCollections can be nested, and nested
        collections must also be instances of SyncedCollection so that
        synchronization occurs even when nested structures are modified.
        Recreating the full nested structure every time data is reloaded from
        file is highly inefficient, so this method performs an in-place update
        that only changes entries that need to be changed.

        Parameters
        ----------
        data : collections.abc.Mapping, optional
            The data to be assigned to this dict. If ``None``, the data is left
            unchanged (Default value = None).
        _validate : bool, optional
            If True, the data will not be validated (Default value = False).

        """
        if data is None:
            # If no data is passed, take no action.
            pass
        elif _mapping_resolver.get_type(data) == "MAPPING":
            with self._suspend_sync:
                for key, new_value in data.items():
                    try:
                        # The most common usage of SyncedCollections is with a
                        # single object referencing an underlying resource at a
                        # time, so we should almost always find that elements
                        # of data are already contained in self._data, so EAFP
                        # is the best choice for performance.
                        existing = self._data[key]
                    except KeyError:
                        # If the item wasn't present at all, we can simply
                        # assign it.
                        if not _validate:
                            self._validate({key: new_value})
                        self._data[key] = self._from_base(new_value, parent=self)
                    else:
                        if new_value == existing:
                            continue
                        if _sc_resolver.get_type(existing) == "SYNCEDCOLLECTION":
                            try:
                                existing._update(new_value)
                                continue
                            except ValueError:
                                pass

                        # Fall through if the new value is not identical to the
                        # existing value and
                        #     1) The existing value is not a SyncedCollection
                        #        (in which case we would have tried to update it), OR
                        #     2) The existing value is a SyncedCollection, but
                        #       the new value is not a compatible type for _update.
                        if not _validate:
                            self._validate({key: new_value})
                        self._data[key] = self._from_base(new_value, parent=self)

                to_remove = [key for key in self._data if key not in data]
                for key in to_remove:
                    del self._data[key]
        else:
            raise ValueError(
                "Unsupported type: {}. The data must be a mapping or None.".format(
                    type(data)
                )
            )

    def __setitem__(self, key, value):
        self._validate({key: value})
        with self._load_and_save, self._suspend_sync:
            self._data[key] = self._from_base(value, parent=self)

    def reset(self, data):
        """Update the instance with new data.

        Parameters
        ----------
        data : mapping
            Data to update the instance.

        Raises
        ------
        ValueError
            If the data is not a mapping.

        """
        if _mapping_resolver.get_type(data) == "MAPPING":
            self._update(data)
            with self._thread_lock:
                self._save()
        else:
            raise ValueError(
                "Unsupported type: {}. The data must be a mapping or None.".format(
                    type(data)
                )
            )

    def keys(self):  # noqa: D102
        self._load()
        return self._data.keys()

    def values(self):  # noqa: D102
        self._load()
        return self._to_base().values()

    def items(self):  # noqa: D102
        self._load()
        return self._to_base().items()

    def get(self, key, default=None):  # noqa: D102
        self._load()
        return self._data.get(key, default)

    def pop(self, key, default=None):  # noqa: D102
        with self._load_and_save:
            ret = self._data.pop(key, default)
        return ret

    def popitem(self):  # noqa: D102
        with self._load_and_save:
            ret = self._data.popitem()
        return ret

    def clear(self):  # noqa: D102
        self._data = {}
        with self._thread_lock:
            self._save()

    def update(self, other=None, **kwargs):  # noqa: D102
        if other is not None:
            # Convert sequence of key, value pairs to dict before validation
            if _mapping_resolver.get_type(other) != "MAPPING":
                other = dict(other)
        else:
            other = {}

        with self._load_and_save:
            # The order here is important to ensure that the promised sequence of
            # overrides is obeyed: kwargs > other > existing data.
            self._update({**self._data, **other, **kwargs})

    def setdefault(self, key, default=None):  # noqa: D102
        with self._load_and_save:
            if key in self._data:
                ret = self._data[key]
            else:
                self._validate({key: default})
                with self._suspend_sync:
                    ret = self._data[key] = self._from_base(default, parent=self)
        return ret



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Define synced_collections version."""


__version__ = "1.0.0"



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implements a MongoDB :class:`~.SyncedCollection` backend."""
from .. import SyncedCollection, SyncedDict, SyncedList
from ..validators import json_format_validator, require_string_key

try:
    import bson

    MONGO = True
except ImportError:
    MONGO = False


class MongoDBCollection(SyncedCollection):
    r"""A :class:`~.SyncedCollection` that synchronizes with a MongoDB document.

    In MongoDB, a database is composed of multiple MongoDB **collections**, which
    are analogous to tables in SQL databases but do not enforce a schema like
    in relational databases. In turn, collections are composed of **documents**,
    which are analogous to rows in a table but are much more flexible, storing
    any valid JSON object in a JSON-like encoded format known as BSON
    ("binary JSON").

    Each :class:`~.MongoDBCollection` can be represented as a MongoDB document,
    so this backend stores the :class:`~.MongoDBCollection` as a single
    document within the collection provided by the user. The document is
    identified by a unique key provided by the user.

    **Thread safety**

    The :class:`MongoDBCollection` is not thread-safe.

    Parameters
    ----------
    collection : :class:`pymongo.collection.Collection`
        The MongoDB client in which to store data.
    uid : dict
        The unique key-value mapping added to the data and stored in the document
        so that it is uniquely identifiable in the MongoDB collection. The key
        "data" is reserved and may not be part of this uid.
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    The user is responsible for providing a unique id such that there are no
    possible collisions between different :class:`~.MongoDBCollection` instances
    stored in the same MongoDB collection. Failure to do so may result in data
    corruption if multiple documents are found to be apparently associated with
    a given ``uid``.

    """

    _backend = __name__  # type: ignore

    # MongoDB uses BSON, which is not exactly JSON but is close enough that
    # JSON-validation is reasonably appropriate. we could generalize this to do
    # proper BSON validation if we find that the discrepancies (for instance, the
    # supported integer data types differ) are too severe.
    _validators = (json_format_validator,)

    def __init__(self, collection=None, uid=None, parent=None, *args, **kwargs):
        super().__init__(parent=parent, **kwargs)
        if not MONGO:
            raise RuntimeError(
                "The PyMongo package must be installed to use the MongoDBCollection."
            )

        self._collection = collection
        if uid is not None and "data" in uid:
            raise ValueError("The key 'data' may not be part of the uid.")
        self._uid = uid

    def _load_from_resource(self):
        """Load the data from a MongoDB document.

        Returns
        -------
        Collection or None
            An equivalent unsynced collection satisfying :meth:`~.is_base_type` that
            contains the data in the MongoDB database. Will return None if no data
            was found in the database.

        """
        blob = self._collection.find_one(self._uid)
        return blob["data"] if blob is not None else None

    def _save_to_resource(self):
        """Write the data to a MongoDB document."""
        data = self._to_base()
        data_to_insert = {**self._uid, "data": data}
        try:
            self._collection.replace_one(self._uid, data_to_insert, True)
        except bson.errors.InvalidDocument as err:
            raise TypeError(str(err))

    @property
    def collection(self):
        """pymongo.collection.Collection: Get the collection being synced to."""
        return self._collection

    @property
    def uid(self):  # noqa: D401
        """dict: Get the unique mapping used to identify this collection."""
        return self._uid

    def __deepcopy__(self, memo):
        # The underlying MongoDB collection cannot be deepcopied.
        raise TypeError("MongoDBCollection does not support deepcopying.")


class MongoDBDict(MongoDBCollection, SyncedDict):
    r"""A dict-like data structure that synchronizes with a document in a MongoDB collection.

    Examples
    --------
    >>> doc = MongoDBDict('data')
    >>> doc['foo'] = "bar"
    >>> assert doc['foo'] == "bar"
    >>> assert 'foo' in doc
    >>> del doc['foo']
    >>> doc['foo'] = dict(bar=True)
    >>> doc
    {'foo': {'bar': True}}

    Parameters
    ----------
    collection : pymongo.collection.Collection, optional
        A :class:`pymongo.collection.Collection` instance (Default value = None).
    uid : dict, optional
        The unique key-value mapping identifying the collection (Default value = None).
    data : non-str :class:`collections.abc.Mapping`, optional
        The initial data passed to :class:`MongoDBDict`. If ``None``, defaults to
        ``{}`` (Default value = None).
    parent : MongoDBCollection, optional
        A parent instance of :class:`MongoDBCollection` or ``None``. If ``None``,
        the collection owns its own data (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    While the :class:`MongoDBDict` object behaves like a :class:`dict`, there are
    important distinctions to remember. In particular, because operations are
    reflected as changes to an underlying database, copying a
    :class:`MongoDBDict` instance may exhibit unexpected behavior. If a true
    copy is required, you should use the call operator to get a dictionary
    representation, and if necessary construct a new :class:`MongoDBDict`
    instance.

    """

    _validators = (require_string_key,)

    def __init__(
        self, collection=None, uid=None, data=None, parent=None, *args, **kwargs
    ):
        super().__init__(
            collection=collection, uid=uid, data=data, parent=parent, *args, **kwargs
        )


class MongoDBList(MongoDBCollection, SyncedList):
    r"""A list-like data structure that synchronizes with a document in a MongoDB collection.

    Only non-string sequences are supported by this class.

    Examples
    --------
    >>> synced_list = MongoDBList('data')
    >>> synced_list.append("bar")
    >>> assert synced_list[0] == "bar"
    >>> assert len(synced_list) == 1
    >>> del synced_list[0]

    Parameters
    ----------
    collection : pymongo.collection.Collection, optional
        A :class:`pymongo.collection.Collection` instance (Default value = None).
    uid : dict, optional
        The unique key-value mapping identifying the collection (Default value = None).
    data : non-str :class:`collections.abc.Sequence`, optional
        The initial data passed to :class:`MongoDBList`. If ``None``, defaults to
        ``[]`` (Default value = None).
    parent : MongoDBCollection, optional
        A parent instance of :class:`MongoDBCollection` or ``None``. If ``None``,
        the collection owns its own data (Default value = None).
    \*args :
        Positional arguments forwarded to parent constructors.
    \*\*kwargs :
        Keyword arguments forwarded to parent constructors.

    Warnings
    --------
    While the :class:`MongoDBList` object behaves like a :class:`list`, there are important
    distinctions to remember. In particular, because operations are reflected
    as changes to an underlying database, copying a :class:`MongoDBList` instance may
    exhibit unexpected behavior. If a true copy is required, you should use the
    call operator to get a dictionary representation, and if necessary
    construct a new :class:`MongoDBList` instance.

    """

    def __init__(
        self, collection=None, uid=None, data=None, parent=None, *args, **kwargs
    ):
        super().__init__(
            collection=collection, uid=uid, data=data, parent=parent, *args, **kwargs
        )



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import pytest
from synced_collection_test import SyncedDictTest, SyncedListTest

from synced_collections.backends.collection_mongodb import MongoDBDict, MongoDBList

try:
    import pymongo

    try:
        # Test the mongodb server. Set a short timeout so that tests don't
        # appear to hang while waiting for a connection.
        mongo_client = pymongo.MongoClient(serverSelectionTimeoutMS=1000)
        tmp_collection = mongo_client["test_db"]["test"]
        tmp_collection.insert_one({"test": "0"})
        ret = tmp_collection.find_one({"test": "0"})
        if ret["test"] != "0":
            raise RuntimeError("MongoDB access check failed.")
        tmp_collection.drop()
        PYMONGO = True
    except (pymongo.errors.ServerSelectionTimeoutError, RuntimeError):
        PYMONGO = False
except ImportError:
    PYMONGO = False


try:
    import numpy

    NUMPY = True

    from synced_collection_test import NUMPY_INT_TYPES, NUMPY_SHAPES

    # BSON does not support >8-byte ints. We remove larger types since some are
    # architecture-dependent.
    NUMPY_INT_TYPES = tuple(
        dtype
        for dtype in NUMPY_INT_TYPES
        if issubclass(dtype, numpy.number)
        and numpy.log2(numpy.iinfo(dtype).max) / 8 < 8
    )
except ImportError:
    NUMPY = False

    NUMPY_INT_TYPES = ()
    NUMPY_SHAPES = ()


class MongoDBCollectionTest:
    _uid = {"MongoDBCollection::name": "test"}

    def store(self, synced_collection, data):
        data_to_insert = {**synced_collection.uid, "data": data}
        synced_collection.collection.replace_one(synced_collection.uid, data_to_insert)

    @pytest.fixture
    def synced_collection(self, request):
        yield self._collection_type(
            uid=self._uid, collection=mongo_client.test_db.test_dict
        )
        mongo_client.test_db.test_dict.drop()

    @pytest.fixture
    def synced_collection_positional(self):
        """Fixture that initializes the object using positional arguments."""
        yield self._collection_type(mongo_client.test_db.test_dict, self._uid)
        mongo_client.test_db.test_dict.drop()

    def test_uid(self, synced_collection):
        assert synced_collection.uid == self._uid

    @pytest.mark.parametrize("dtype", NUMPY_INT_TYPES)
    @pytest.mark.parametrize("shape", NUMPY_SHAPES)
    def test_set_get_numpy_int_data(self, synced_collection, dtype, shape):
        """Override parent test to use the subset of int types."""
        super().test_set_get_numpy_int_data(synced_collection, dtype, shape)


@pytest.mark.skipif(
    not PYMONGO, reason="test requires the pymongo package and mongodb server"
)
class TestMongoDBDict(MongoDBCollectionTest, SyncedDictTest):
    _collection_type = MongoDBDict


@pytest.mark.skipif(
    not PYMONGO, reason="test requires the pymongo package and mongodb server"
)
class TestMongoDBList(MongoDBCollectionTest, SyncedListTest):
    _collection_type = MongoDBList

    @pytest.mark.parametrize("dtype", NUMPY_INT_TYPES)
    @pytest.mark.parametrize("shape", (None, (1,), (2,)))
    def test_reset_numpy_int_data(self, synced_collection, dtype, shape):
        """Override parent test to use the subset of int types."""
        super().test_reset_numpy_int_data(synced_collection, dtype, shape)



# Copyright (c) 2023 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implement the caching feature to SyncedCollection API."""
import logging
import pickle
import uuid
from collections.abc import MutableMapping

logger = logging.getLogger(__name__)


def get_cache():
    """Return the cache.

    This method returns an instance of :class:`~RedisCache` if a Redis server
    is available, or otherwise an instance of :class:`dict` for an in-memory
    cache.

    Returns
    -------
    cache
        An instance of :class:`~_RedisCache` if redis-server is available,
        otherwise a dict.

    """
    try:
        import redis

        REDIS = True
    except ImportError as error:
        logger.debug(str(error))
        REDIS = False
    if REDIS:
        try:
            # try to connect to server
            cache = redis.Redis()
            test_key = str(uuid.uuid4())
            cache.set(test_key, 0)
            if cache.get(test_key) != b"0":  # Redis stores data as bytes
                raise RuntimeError("Cache access check failed.")
            cache.delete(test_key)
            logger.info("Using Redis cache.")
            return _RedisCache(cache)
        except (redis.exceptions.ConnectionError, RuntimeError) as error:
            logger.debug(str(error))
    logger.info("Redis not available.")
    return {}


class _RedisCache(MutableMapping):
    """Redis-based cache.

    Redis restricts the types of data it can handle to bytes, strings, or
    numbers, and it always returns responses as bytes. The RedisCache is a
    :class:`~collections.abc.MutableMapping` that provides a convenient wrapper
    around instances of :class:`redis.Redis`, handling conversions to and from
    the appropriate data types.
    """

    def __init__(self, client):
        self._client = client

    def __setitem__(self, key, value):
        self._client[key] = pickle.dumps(value)

    def __getitem__(self, key):
        return pickle.loads(self._client[key])

    def __delitem__(self, key):
        self._client.delete(key)

    def __contains__(self, key):
        return key in self._client

    def __iter__(self):
        for key in self._client.keys():
            yield key.decode()

    def __len__(self):
        return len(self._client.keys())
